///////////////////////////////////////////////////////////////////////////////
//                        DecayCalc Class
//    This class handles calculation of the slepton flavor-violating decays.
//  The class is responsible both for establishing the slepton mass hierarchy 
//  and recalculating decay widths for all decays affected by flavor-violating 
//  sleptons are recomputed - both the decays of sleptons and sneutrinos 
//  themselves, and the decays of other SUSY particles into sleptons and 
//  sneutrinos.  The class maintains a pointer to the DecayFormat object which 
//  instantiates it, and all actual output to the intermediate file is handled 
//  through calls to DecayFormat methods.  The class contains a 
//  ThreeBodyIntegrals object to calculate the integrals over phase space 
//  present three body light slepton decays.
//    The class has three major types of methods:
//
//  Output[...]Decays     Outputs decays of the given particles -
//                        Neutralinos, Charginos, Sleptons, Sneutrinos, 
//                        and Higgs bosons.  Each method loops over all
//                        possible mass indices for both parent and product
//                        particles.  We assume that CP violation is absent,
//                        and thus set equal charge-conjugate decays.
//
//  Calc[...]             Calculates the decay width for the given decay
//                        mode.  Formulas correspond to those given in
//                        Appendix D.
//                        The index convention for mass eigenstates is:
//                          i - mass eigenstate of decaying particle
//                          j - mass eigenstate of the first decay product,
//                              always chosen to be a SUSY particle
//                          k - mass eigenstate of the second decay product,
//                              in all cases but the Higgs decays this is 
//                              taken to be a/the non-SUSY decay product
//                          m - the third decay product for three-body decays
//
//  Alpha, Beta1, etc.    Flavor-violating coupling constants from the 
//                        Lagrangian.  These functions are chosen to be real
//                        (by demanding all mixing matrices are real) for ease 
//                        of computation.  For more details reference 
//                        Appendix B of the manual.
///////////////////////////////////////////////////////////////////////////////

#include <DecayCalc.h>
#include <DecayFormat.h>
#include <ThreeBodyIntegrals.h>
#include <flavoursoft.h>
#include <linalg.h>
#include <fstream>
#include <iomanip>

//Constructor for the DecayCalc class
//Stores pointers to the passed DecayFormat object (which should be the same 
//DecayFormat object from which DecayCalc is instantiated) and instantiates 
//a ThreeBodyIntegrals object.  The constructor also takes in the 6x6 slepton 
//mass matrix and 3x3 sneutrino mass matrix and diagonalizes them, generating 
//vectors of slepton and sneutrino masses and corresponding mixing matrices.
DecayCalc::DecayCalc(DecayFormat *output)
: slRot(6,6), snuRot(3,3), slMass(6), snuMass(3) {

  //Assign pointers
  decayOut = output;
  model = decayOut->getModel();
  integrals = new ThreeBodyIntegrals();

  //Take slepton/sneutrino mass vector and mixing matrix from the model object
  slMass = model->displayFlavourPhysical().msE;
  snuMass = model->displayFlavourPhysical().msNu;
  slRot = model->displayFlavourPhysical().eSqMix.transpose();
  snuRot = model->displayFlavourPhysical().nuSqMix.transpose();
}

// Evaluates [1]: \chi_i -> \tilde{l}_j \bar{l}_k
double DecayCalc::CalcChi0SlepLepBar(int i, int j, int k) {
  double m_i, m_j, m_k, width, beta1, beta2, lambda;

  //Assign masses
  m_i = model->displayPhys().mneut.display(i);
  m_j = slMass(j);
  m_k = LepMass(k);

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j) + fabs(m_k)) return 0;

  lambda = Lambda(sqr(m_i), sqr(m_j), sqr(m_k));
  beta1 = Beta1(i,j,k);
  beta2 = Beta2(i,j,k);

  width = (sqr(beta1) + sqr(beta2)) * (sqr(m_i) + sqr(m_k) - sqr(m_j))
         + 4 * beta1 * beta2 * m_i * m_k;

  width *= 1./(32.*PI * pow(fabs(m_i),3)) * sqrt(lambda);

  return width;
}

// Evaluates [2]: \chi_k -> \tilde{\nu}_i \bar{\nu}_j
double DecayCalc::CalcChi0SnuNuBar(int i,int j,int k) {
  double m_i, m_j, width, alpha;

  //Assign masses
  m_i = model->displayPhys().mneut.display(i);
  m_j = snuMass(j);

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j)) return 0;

  alpha = Alpha(i,j,k);

  width = sqr(alpha) * sqr(sqr(m_i) - sqr(m_j));
  width *= 1./(32.*PI * pow(fabs(m_i),3));

  return width;
}

//Calculates all neutralino decays and passes them to the DecayFormat object 
//for output.  Only positive-definite decays are output; decays with widths 
//of zero are discarded.
void DecayCalc::OutputNeutralinoDecays() {
  double width;
  int i,j,k;

  for(i = 1; i <= 4; i++) {
    for(j = 1; j <= 6; j++) {
      for(k = 1; k <= 3; k++) {
	//[1]: \chi_i -> \tilde{l}_j \bar{l}_k
        width = CalcChi0SlepLepBar(i,j,k);
        if(width > 0) {
          decayOut->TwoBodyDecay(width,NEUTRALINO,i,SLEPTON,j,ANTILEPTON,k);
          decayOut->TwoBodyDecay(width,NEUTRALINO,i,ANTISLEPTON,j,LEPTON,k);
        }
      }
    }
    for(j = 1; j <= 3; j++) {
      for(k = 1; k <= 3; k++) {
	//[2]: \chi_k -> \tilde{\nu}_i \bar{\nu}_j
        width = CalcChi0SnuNuBar(i,j,k);
        if (width > 0) {
          decayOut->TwoBodyDecay(width,NEUTRALINO,i,SNEUTRINO,j,
                                 ANTINEUTRINO,k);
          decayOut->TwoBodyDecay(width,NEUTRALINO,i,ANTISNEUTRINO,j,
                                 NEUTRINO,k);
        }
      }
    }
  }
}

// Evaluates [3]: \tilde{\chi}^+_k -> \nu_j \tilde l^*_i
double DecayCalc::CalcChiMinusSlepNuBar(int i,int j,int k) {
  double m_i, m_j, gamma, width;

  //Assign masses
  m_i = model->displayPhys().mch.display(i);
  m_j = slMass(j);

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j)) return 0;

  gamma = Gamma(i,j,k);

  width = 1./(32.*PI * pow(fabs(m_i),3)) * sqr(gamma) * sqr(sqr(m_i)-sqr(m_j));
  return width;
}

// Evaluates [4]: \tilde{\chi}^+_k -> \tilde{\nu}_i \bar{l}_j
double DecayCalc::CalcChiPlusSnuLepBar(int i,int j,int k) {
  double m_i, m_j, m_k, delta1, delta2, lambda, width;

  //Assign masses
  m_i = model->displayPhys().mch.display(i);
  m_j = snuMass(j);
  m_k = LepMass(k);

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j) + fabs(m_k)) return 0;

  lambda = Lambda(sqr(m_i), sqr(m_j), sqr(m_k));
  delta1 = Delta1(i,j,k);
  delta2 = Delta2(i,j,k);

  width = (sqr(delta1) + sqr(delta2)) * (sqr(m_i) - sqr(m_j) + sqr(m_k))
              + 4 * m_i * m_k * delta1 * delta2;
  width *= 1./(32.*PI * pow(fabs(m_i),3)) * sqrt(lambda);

  return width;
}

//Calculates all chargino decays and passes them to the DecayFormat object 
//for output.  Only positive-definite decays are output; decays with widths 
//of zero are discarded.
void DecayCalc::OutputCharginoDecays() {
  double width;
  int i,j,k;

  for(i = 1; i <= 2; i++) {
    for(j = 1; j <= 3; j++) {
      for(k = 1; k <= 3; k++) {
	//[3]: \tilde{\chi}^+_k -> \nu_j \tilde l^*_i
        width = CalcChiPlusSnuLepBar(i,j,k);
        if (width > 0) {
          decayOut->TwoBodyDecay(width,CHARGINOPLUS,i,SNEUTRINO,j,
                                 ANTILEPTON,k);
          decayOut->TwoBodyDecay(width,CHARGINOMINUS,i,ANTISNEUTRINO,j,
                                LEPTON,k);
        }
      }
    }
    for(j = 1; j <= 6; j++) {
      for(k = 1; k <= 3; k++) {
	//[4]: \tilde{\chi}^+_k -> \tilde{\nu}_i \bar{l}_j
        width = CalcChiMinusSlepNuBar(i,j,k);
        if (width > 0) {
          decayOut->TwoBodyDecay(width,CHARGINOMINUS,i,SLEPTON,j,
                                ANTINEUTRINO,k);
          decayOut->TwoBodyDecay(width,CHARGINOPLUS,i,ANTISLEPTON,j,
                                NEUTRINO,k);
        }
      }
    }
  }
}

// Evaluates [5]: \tilde{l}_i -> \chi_j^0 l_k
double DecayCalc::CalcSlepChi0Lep(int i,int j,int k) {
  double m_i, m_j, m_k, width, lambda, beta1, beta2;

  //Assign masses
  m_i = slMass(i);
  m_j = model->displayPhys().mneut.display(j);
  m_k = LepMass(k);

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j) + fabs(m_k)) return 0;

  lambda = Lambda(sqr(m_i), sqr(m_j), sqr(m_k));
  beta1 = Beta1(j,i,k);
  beta2 = Beta2(j,i,k);

  width = (sqr(beta1) + sqr(beta2)) * (sqr(m_i) - sqr(m_j) - sqr(m_k))
          - 4 * beta1 * beta2 * m_k * m_j;

  width *= 1./(16.*PI * pow(m_i,3)) * sqrt(lambda);

  return width;
}

// Evaluates [6]: \tilde{l}_i -> \chi^-_j \nu_k
double DecayCalc::CalcSlepChiMinusNu(int i,int j,int k) {
  double m_i, m_j, gamma, width;

  //Assign masses
  m_i = slMass(i);
  m_j = model->displayPhys().mch.display(j);

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j)) return 0;

  gamma = Gamma(j,i,k);

  width = 1./(16.*PI * pow(fabs(m_i),3)) * sqr(gamma) * sqr(sqr(m_i)-sqr(m_j));
  return width;
}

// Evaluates [7]: \tilde{l}_i -> \tilde{l}_j h^0
double DecayCalc::CalcSlepSlepHiggs(int i,int j) {
  double m_i , m_j, m_h, sigma2, width, lambda;

  //Assign masses
  m_i = slMass(i);
  m_j = slMass(j);
  m_h = model->displayPhys().mhiggs.display(1);

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j) + fabs(m_h)) return 0;

  lambda = Lambda(sqr(m_i),sqr(m_j),sqr(m_h));
  sigma2 = Sigma2(j,i);

  width = sqr(sigma2) / (16.*PI * pow(m_i,3)) * sqrt(lambda);

  return width;
}

// Evaluates [8]: \tilde{l}_i -> \tilde{l}_j Z
double DecayCalc::CalcSlepSlepZ(int i,int j) {
  double m_i, m_j, m_z, zeta2, width, lambda;

  //Assign masses
  m_i = slMass(i);
  m_j = slMass(j);
  m_z = model->displayMz();

  //Restrict to kinematically allowed decays
  if(fabs(m_i) <= fabs(m_j) + fabs(m_z)) return 0;

  lambda = Lambda(sqr(m_i), sqr(m_j), sqr(m_z));
  zeta2 = Zeta2(j,i);

  width = sqr(zeta2) * pow(lambda,1.5) / (16.* PI * sqr(m_z) * pow(m_i,3));

  return width;
}

// Evaluates [9]: \tilde{l}_i -> \tilde{\nu}_j W^-
double DecayCalc::CalcSlepSnuWMinus(int i,int j) {
  double m_i, m_j, m_W, width, zeta1, lambda;

  //Assign masses
  m_i = slMass(i);
  m_j = snuMass(j);
  m_W = model->displayMw();

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j) + fabs(m_W)) return 0;

  lambda = Lambda(sqr(m_i), sqr(m_j), sqr(m_W));
  zeta1 = Zeta1(j,i);

  width = sqr(zeta1) * pow(lambda,1.5) / (16. * PI * sqr(m_W) * pow(m_i,3.0));

  return width;
}

//////////////////////////////////////////////////////////////////////////////

// Evaluates [17]: \tilde{l}_i -> \tilde{l}_j + l_k + \overline{l}_m
// The 'standard' process for this decay is through an off-shell neutrlino, 
// but in the flavor violating case there are contributions from an off-shell
// Higgs or Z boson diagram when k = m
double DecayCalc::CalcSlepSlepLepLepBar(int i,int j,int k,int m) {
  double m0, width, widthNZ, widthNH, widthZZ, widthZH, widthHH, cZH, cHH;
  int a, b;
  DoubleVector mass(5), cNN(6), cZZ(2), cNZ(4), cNH(2);

  //Assign masses for parent and daughter particles
  // mass(4) and mass(5) are the masses of internal off-shell
  // particles, to be set individually for each mode
  m0 = slMass(i);
  mass(1) = slMass(j);
  mass(2) = LepMass(k);
  mass(3) = LepMass(m);

  //Restrict to kinematically allowed decays
  if(m0 <= fabs(mass(1)) + fabs(mass(2)) + fabs(mass(3))) return 0;

  //Restrict decays to sleptons which cannot decay to any on-shell neutralino
  //(assumes lightest neutralino is lighter than light chargino)
  //Sleptons which can decay to an on-shell neutralino will undergo two 
  //separate two-body decays.
  if(m0 >= fabs(model->displayPhys().mneut.display(1)) + MELECTRON) 
    return 0;

  width = 0;

  //Start the calculation by calculating the contribution of the neutralino
  //intermediate to the decay width since this contribution is present for
  //all lepton combinations

  ///////////////////////////// Chi0Chi0 //////////////////////////////////////

  //Loop over the neutralino eigenstate
  for(a = 1; a <= 4; a++) {
    mass(4) = model->displayPhys().mneut.display(a);
    for(b = 1; b <= 4; b++) {
      mass(5) = model->displayPhys().mneut.display(b);
      cNN(1) = IntegralCoefficient(i,j,k,m,a,b,1,1,1,1);
      cNN(2) = IntegralCoefficient(i,j,k,m,a,b,1,2,1,2);
      cNN(3) = 2 * IntegralCoefficient(i,j,k,m,a,b,1,1,1,2);
      cNN(4) = - 2 * IntegralCoefficient(i,j,k,m,a,b,1,1,2,1);
      cNN(5) = - 2 * IntegralCoefficient(i,j,k,m,a,b,1,2,2,1);
      cNN(6) = - 2 * IntegralCoefficient(i,j,k,m,a,b,1,1,2,2);
      integrals->setMasses(mass,m0);
      width += integrals->evaluateNeutralinoIntegrals(cNN);
    }
  }

  //Scale width by appropriate factors
  width *= m0 / (512 * PI * PI * PI);

  // If outgoing leptons are the same generation, there are standard and
  // interference terms from additional Z and Higgs boson modes
  if(k == m) {

    /////////////////////////////// ZZ ////////////////////////////////////////

    mass(4) = model->displayMz();
    mass(5) = mass(4);
    cZZ(1) = ( sqr(LeftZCoupling(LEPTON)) + sqr(RightZCoupling(LEPTON)) ) 
      * sqr(Zeta2(j,i));
    cZZ(2) = 2. * LeftZCoupling(LEPTON) * RightZCoupling(LEPTON) 
      * sqr(Zeta2(j,i));
    integrals->setMasses(mass,m0);
    widthZZ = integrals->evaluateGaugeIntegrals(cZZ);

    //Scale width by appropriate factors
    widthZZ *= m0 / (128 * PI * PI * PI);

    //////////////////////////// HiggsHiggs ///////////////////////////////////

    widthHH = 0;

    //h0h0 contribution
    mass(4) = model->displayPhys().mhiggs.display(1);
    mass(5) = mass(4);
    cHH = sqr(Sigma2(j,i) * HiggsCoupling(SMHIGGS,LEPTON,k));
    integrals->setMasses(mass,m0);
    widthHH += integrals->evaluateScalarIntegral(cHH);

    //H0H0 contribution
    mass(4) = model->displayPhys().mhiggs.display(3);
    mass(5) = mass(4);
    cHH = sqr(Sigma3(j,i) * HiggsCoupling(HEAVYHIGGS,LEPTON,k));
    integrals->setMasses(mass,m0);
    widthHH += integrals->evaluateScalarIntegral(cHH);

    //A0A0 contribution
    mass(4) = model->displayPhys().mhiggs.display(2);
    mass(5) = mass(4);
    cHH = sqr(Sigma4(j,i) * HiggsCoupling(CPODDHIGGS,LEPTON,k));
    integrals->setMasses(mass,m0);
    widthHH += integrals->evaluatePseudoscalarIntegral(cHH);

    //G0G0 contribution
    mass(4) = model->displayMz();
    mass(5) = mass(4);
    cHH = sqr(Sigma5(j,i) * HiggsCoupling(G0HIGGS,LEPTON,k));
    integrals->setMasses(mass,m0);
    widthHH += integrals->evaluatePseudoscalarIntegral(cHH);

    //HiggsHiggs interference terms

    //h0H0 contribution
    mass(4) = model->displayPhys().mhiggs.display(1);
    mass(5) = model->displayPhys().mhiggs.display(3);
    cHH = Sigma2(j,i) * Sigma3(j,i) * HiggsCoupling(SMHIGGS,LEPTON,k)
      * HiggsCoupling(HEAVYHIGGS,LEPTON,k);
    integrals->setMasses(mass,m0);
    widthHH += 2 * integrals->evaluateScalarScalarCrossIntegral(cHH);

    //A0G0 contribution
    mass(4) = model->displayPhys().mhiggs.display(2);
    mass(5) = model->displayMz();
    cHH = Sigma4(j,i) * Sigma5(j,i) * HiggsCoupling(CPODDHIGGS,LEPTON,k)
      * HiggsCoupling(G0HIGGS,LEPTON,k);
    integrals->setMasses(mass,m0);
    widthHH += 2*integrals->evaluatePseudoscalarPseudoscalarCrossIntegral(cHH);

    //Scale width by appropriate factors
    widthHH *= 1.0 / (128 * PI * PI * PI * m0);

    /////////////////////////////// ZChi0/ ////////////////////////////////////

    widthNZ=0;

    double L = LeftZCoupling(LEPTON) * Zeta2(j,i);
    double R = RightZCoupling(LEPTON) * Zeta2(j,i);

    //Loop over the neutralino eigenstate
    for(b = 1; b <= 4; b++) {
      mass(4) = model->displayMz();
      mass(5) = model->displayPhys().mneut.display(b);
      cNZ(1) = L * Beta1(b,i,k)*Beta1(b,j,m) + R * Beta2(b,i,k)*Beta2(b,j,m);
      cNZ(2) = L * Beta1(b,i,k)*Beta2(b,j,m) + R * Beta2(b,i,k)*Beta1(b,j,m);
      cNZ(3) = R * Beta1(b,i,k)*Beta2(b,j,m) + L * Beta2(b,i,k)*Beta1(b,j,m);
      cNZ(4) = R * Beta1(b,i,k)*Beta1(b,j,m) + L * Beta2(b,i,k)*Beta2(b,j,m);
      integrals->setMasses(mass,m0);
      widthNZ += integrals->evaluateGaugeFermionCrossIntegrals(cNZ);
    }

    //Scale width by appropriate factors
    widthNZ *= - m0 / (256 * PI * PI * PI);

    /////////////////////////////// HiggsChi0 /////////////////////////////////

    widthNH = 0;

    //Chi0h0 contribution
    for(b = 1; b <= 4; b++) {
      mass(4) = model-> displayPhys().mhiggs.display(1);
      mass(5) = model->displayPhys().mneut.display(b);
      cNH(1) = - (Beta1(b,i,k) * Beta2(b,j,m) + Beta2(b,i,k) * Beta1(b,j,m))
      	* Sigma2(j,i) * HiggsCoupling(SMHIGGS,LEPTON,k);
      cNH(2) = - (Beta1(b,i,k) * Beta1(b,j,m) + Beta2(b,i,k) * Beta2(b,j,m))
	* Sigma2(j,i) * HiggsCoupling(SMHIGGS,LEPTON,k);
      integrals->setMasses(mass,m0);
      widthNH += integrals->evaluateScalarFermionCrossIntegrals(cNH);
    }

    //Chi0H0 contribution
    for(b = 1; b <= 4; b++) {
      mass(4) = model->displayPhys().mhiggs.display(3);
      mass(5) = model->displayPhys().mneut.display(b);
      cNH(1) = - (Beta1(b,i,k) * Beta2(b,j,m) + Beta2(b,i,k) * Beta1(b,j,m))
      	* Sigma3(j,i) * HiggsCoupling(HEAVYHIGGS,LEPTON,k);
      cNH(2) = - (Beta1(b,i,k) * Beta1(b,j,m) + Beta2(b,i,k) * Beta2(b,j,m))
	* Sigma3(j,i) * HiggsCoupling(HEAVYHIGGS,LEPTON,k);
      integrals->setMasses(mass,m0);
      widthNH += integrals->evaluateScalarFermionCrossIntegrals(cNH);
    }

    //Chi0A0 contribution
    for(b = 1; b <= 4; b++) {
      mass(4) = model-> displayPhys().mhiggs.display(2);
      mass(5) = model->displayPhys().mneut.display(b);
      cNH(1) = - (Beta1(b,i,k) * Beta2(b,j,m) - Beta2(b,i,k) * Beta1(b,j,m))
      	* Sigma4(j,i) * HiggsCoupling(CPODDHIGGS,LEPTON,k);
      cNH(2) = - (Beta1(b,i,k) * Beta1(b,j,m) - Beta2(b,i,k) * Beta2(b,j,m))
	* Sigma4(j,i) * HiggsCoupling(CPODDHIGGS,LEPTON,k);
      integrals->setMasses(mass,m0);
      widthNH += integrals->evaluatePseudoscalarFermionCrossIntegrals(cNH);
    }

    //Chi0G0 contribution
    for(b = 1; b <= 4; b++) {
      mass(4) = model-> displayMz();
      mass(5) = model->displayPhys().mneut.display(b);
      cNH(1) = - (Beta1(b,i,k) * Beta2(b,j,m) - Beta2(b,i,k) * Beta1(b,j,m))
      	* Sigma5(j,i) * HiggsCoupling(G0HIGGS,LEPTON,k);
      cNH(2) = - (Beta1(b,i,k) * Beta1(b,j,m) - Beta2(b,i,k) * Beta2(b,j,m))
	* Sigma5(j,i) * HiggsCoupling(G0HIGGS,LEPTON,k);
      integrals->setMasses(mass,m0);
      widthNH += integrals->evaluatePseudoscalarFermionCrossIntegrals(cNH);
    }

    //Scale width by appropriate factors
    widthNH *= 1.0 / (256 * PI * PI * PI);

    ///////////////////////////////// ZHiggs //////////////////////////////////

    widthZH = 0;

    //Zh0 contribution
    mass(4) = model->displayMz();
    mass(5) = model->displayPhys().mhiggs.display(1);
    cZH = Sigma2(j,i) * Zeta2(j,i) * HiggsCoupling(SMHIGGS,LEPTON,k)
      * ( LeftZCoupling(LEPTON) + RightZCoupling(LEPTON) );
    integrals->setMasses(mass,m0);
    widthZH += integrals->evaluateScalarGaugeCrossIntegral(cZH);

    //ZH0 contribution
    mass(4) = model->displayMz();
    mass(5) = model->displayPhys().mhiggs.display(3);
    cZH = Sigma3(j,i) * Zeta2(j,i) * HiggsCoupling(HEAVYHIGGS,LEPTON,k)
      * ( LeftZCoupling(LEPTON) + RightZCoupling(LEPTON) );
    integrals->setMasses(mass,m0);
    widthZH += integrals->evaluateScalarGaugeCrossIntegral(cZH);

    //ZA0 contribution
    mass(4) = model->displayMz();
    mass(5) = model->displayPhys().mhiggs.display(2);
    cZH = - Sigma4(j,i) * Zeta2(j,i) * HiggsCoupling(CPODDHIGGS,LEPTON,k)
      * ( LeftZCoupling(LEPTON) - RightZCoupling(LEPTON) );
    integrals->setMasses(mass,m0);
    widthZH += integrals->evaluateA0GaugeCrossIntegral(cZH);

    //ZG0 contribution
    mass(4) = model->displayMz();
    mass(5) = mass(4);
    cZH = - Sigma5(j,i) * Zeta2(j,i) * HiggsCoupling(G0HIGGS,LEPTON,k)
      * ( LeftZCoupling(LEPTON) - RightZCoupling(LEPTON) );
    integrals->setMasses(mass,m0);
    widthZH += integrals->evaluateG0GaugeCrossIntegral(cZH);

    //Scale width by appropriate factors
    widthZH *= 1.0 / (128 * PI * PI * PI);

    /////////////////////////////// Total /////////////////////////////////////

    //Sum all contributions
    width += widthZZ + widthHH + 2 * widthNZ + 2 * widthNH + 2 * widthZH;
  }

  return width;
}

// Evaluates [18]: \tilde{l}_i -> \tilde{l}_j^* + l_k + l_m
// This decay is mediated by an intermediate off-shell neutralino, and is 
// allowed by the majorana nature of said neutralino.  In the flavor-violating 
// case, either lepton may be emitted from either vertex, producing an 
// interference term
double DecayCalc::CalcSlepSlepBarLepLep(int i,int j,int k,int m) {
  double temp, m0, width, width11, width22, width12;
  int a,b;
  DoubleVector mass(5),c(8);

  //Assign masses for parent and daughter particles
  // mass(4) and mass(5) are the masses of internal off-shell
  // particles, to be set individually for each mode
  m0 = slMass(i);
  mass(1) = slMass(j);
  mass(2) = LepMass(k);
  mass(3) = LepMass(m);

  //Restrict to kinematically allowed decays
  if(m0 <= fabs(mass(1)) + fabs(mass(2)) + fabs(mass(3))) return 0;

  //Restrict decays to sleptons which cannot decay to any on-shell neutralino
  //(assumes lightest neutralino is lighter than light chargino)
  //Sleptons which can decay to an on-shell neutralino will undergo two 
  //separate two-body decays.
  if(m0 >= fabs(model->displayPhys().mneut.display(1)) + MELECTRON) 
    return 0;

  width11 = 0;
  width22 = 0;
  width12 = 0;

  //////////////////////////// "Standard" Constribution ///////////////////////

  //Loop over the neutralino eigenstates
  for(a = 1; a <= 4; a++) {
    mass(4) = model->displayPhys().mneut.display(a);
    for(b = 1; b <= 4; b++) {
      mass(5) = model->displayPhys().mneut.display(b);
      c(1) = IntegralCoefficient(i,j,k,m,a,b,1,2,1,2);
      c(2) = IntegralCoefficient(i,j,k,m,a,b,1,1,1,1);
      c(3) = 2 * IntegralCoefficient(i,j,k,m,a,b,1,2,1,1);
      c(4) = - 2 * IntegralCoefficient(i,j,k,m,a,b,1,2,2,2);
      c(5) = - 2 * IntegralCoefficient(i,j,k,m,a,b,1,1,2,2);
      c(6) = - 2 * IntegralCoefficient(i,j,k,m,a,b,1,2,2,1);
      integrals->setMasses(mass,m0);
      width11 += integrals->evaluateNeutralinoIntegrals(c);
    }
  }

  //Scale width by appropriate factors
  width11 *= m0 / (512 * PI * PI * PI);

  ////////////////////// Interchanged Lepton Constribution ////////////////////

  //This second diagram has exactly the same form as the 'standard' 
  //contribution except with l_k <-> l_m.  The only difference between this
  //and the previous code block is the masses are switched and k <-> m in all
  //coefficients.

  //Switch lepton masses
  temp = mass(2);
  mass(2) = mass(3);
  mass(3) = temp;

  //Loop over the neutralino eigenstates
  for(a = 1; a <= 4; a++) {
    mass(4) = model->displayPhys().mneut.display(a);
    for(b = 1; b <= 4; b++) {
      mass(5) = model->displayPhys().mneut.display(b);
      c(1) = IntegralCoefficient(i,j,m,k,a,b,1,2,1,2);
      c(2) = IntegralCoefficient(i,j,m,k,a,b,1,1,1,1);
      c(3) = 2 * IntegralCoefficient(i,j,m,k,a,b,1,2,1,1);
      c(4) = - 2 * IntegralCoefficient(i,j,m,k,a,b,1,2,2,2);
      c(5) = - 2 * IntegralCoefficient(i,j,m,k,a,b,1,1,2,2);
      c(6) = - 2 * IntegralCoefficient(i,j,m,k,a,b,1,2,2,1);
      integrals->setMasses(mass,m0);
      width22 += integrals->evaluateNeutralinoIntegrals(c);
    }
  }

  //Scale width by appropriate factors
  width22 *= m0 / (512 * PI * PI * PI);

  /////////////////////// Interference Term Constribution /////////////////////

  //Switch lepton masses back to original case
  temp = mass(2);
  mass(2) = mass(3);
  mass(3) = temp;

  //Loop over the neutralino eigenstates
  for(a = 1; a <= 4; a++) {
    mass(4) = model->displayPhys().mneut.display(a);
    for(b = 1; b <= 4; b++) {
      mass(5) = model->displayPhys().mneut.display(b);
      c(1) = CrossCoefficient(i,j,k,m,a,b,1,2,2,1);
      c(2) = - CrossCoefficient(i,j,k,m,a,b,1,2,1,2);
      c(3) = CrossCoefficient(i,j,k,m,a,b,1,2,2,2);
      c(4) = - CrossCoefficient(i,j,k,m,a,b,1,2,1,1);
      c(5) = - CrossCoefficient(i,j,k,m,a,b,1,1,1,2);
      c(6) = CrossCoefficient(i,j,k,m,a,b,1,1,2,1);
      c(7) = - CrossCoefficient(i,j,k,m,a,b,1,1,1,1);
      c(8) = CrossCoefficient(i,j,k,m,a,b,1,1,2,2);
      integrals->setMasses(mass,m0);
      width12 += integrals->evaluateNeutralinoCrossIntegrals(c);
    }
  }

  //Scale width by appropriate factors
  width12 *= m0 / (256 * PI * PI * PI);

  //Sum total contribution
  width = width11 + width22 - 2 * width12;

  //Divide width by two for the case of two identical leptons
  if(k == m) width /= 2;

  return width;
}

// Evaluates [19]: \tilde{l}_i -> \tilde{l}_j + nu_k + nubar_m
// The 'standard' process for this decay is through an off-shell chargino, and 
// in the flavor-violating case there is a contribution from an off-shell Z
// boson diagram when k = m.
double DecayCalc::CalcSlepSlepNuNuBar(int i,int j,int k,int m) {
  double m0, width, c;
  int a,b;
  DoubleVector mass(5);

  //Assign masses for parent and daughter particles
  // mass(4) and mass(5) are the masses of internal off-shell
  // particles, to be set individually for each mode
  m0 = slMass(i);
  mass(1) = slMass(j);
  mass(2) = 0;
  mass(3) = 0;

  //Restrict to kinematically allowed decays
  if(m0 <= mass(1)) return 0;

  //Restrict decays to sleptons which cannot decay to any on-shell neutralino
  //(assumes lightest neutralino is lighter than light chargino)
  //Sleptons which can decay to an on-shell neutralino will undergo two 
  //separate two-body decays.
  if(m0 >= fabs(model->displayPhys().mneut.display(1)) + MELECTRON) return 0;

  width = 0;

  //Start the calculation by calculating the contribution of the neutralino
  //intermediate to the decay width since this contribution is present for
  //all lepton combinations

  /////////////////////////////// ChiChi //////////////////////////////////////

  //Loop over the chargino eigenstate
  for(a = 1; a <= 2; a++) {
    mass(4) = model->displayPhys().mch.display(a);
    for(b = 1; b <= 2; b++) {
      mass(5) = model->displayPhys().mch.display(b);
      c = Gamma(a,i,k) * Gamma(a,j,m) * Gamma(b,i,k) * Gamma(b,j,m);
      integrals->setMasses(mass,m0);
      width += integrals->evaluateCharginoIntegral(c);
    }
  }

  //Scale width by appropriate factors
  width *= m0 / (512 * PI * PI * PI);

  // If outgoing leptons are the same generation, there are standard and
  // interference terms from additional Z and Higgs boson modes
  if(k == m) {

    double cCZ;
    double widthZZ, widthCZ;
    DoubleVector cZZ(2);

    /////////////////////////////// ZZ ////////////////////////////////////////

    mass(4) = model->displayMz();
    mass(5) = mass(4);
    cZZ(1) = sqr(LeftZCoupling(NEUTRINO) * Zeta2(j,i));
    cZZ(2) = 0;
    integrals->setMasses(mass,m0);
    widthZZ = integrals->evaluateGaugeIntegrals(cZZ);

    //Scale width by appropriate factors
    widthZZ *= m0 / (128 * PI * PI * PI);

    /////////////////////////////// ZChi //////////////////////////////////////

    widthCZ=0;

    //Loop over chargino eigenstates
    for(b = 1; b <= 2; b++) {
      mass(4) = model->displayMz();
      mass(5) = model->displayPhys().mch.display(b);
      cCZ = - Gamma(b,i,k) * Gamma(b,j,m) * LeftZCoupling(NEUTRINO) 
      	* Zeta2(j,i);
      integrals->setMasses(mass,m0);
      widthCZ += integrals->evaluateGaugeCharginoCrossIntegral(cCZ);
    }

    //Scale width by appropriate factors
    widthCZ *= m0 / (256 * PI * PI * PI);

    //Sum all widths
    width += widthZZ + 2 * widthCZ;
  }

  return width;
}

// Evaluates [20]: \tilde{l}_i -> \tilde{l}_j^* + q \overline{q}
// In the flavor-violating case there are decays of this form mediated by an
// off-shell Higgs or Z boson.  Only quarks of the same generation are allowed
// by this decay.
double DecayCalc::CalcSlepSlepQQBar(int i,int j,int k,particletype q) {

  //Restrict to quarks
  if(q != UPQUARK && q != DOWNQUARK) {
    cout<<"Error: attempting to calculate quark mode with non-quark."<<endl;
    return 0;
  }

  double m0, width, widthZZ, widthZH, widthHH, cZH, cHH;
  int a,b;
  DoubleVector mass(5), cZZ(2);

  //Assign masses for parent and daughter particles
  // mass(4) and mass(5) are the masses of internal off-shell
  // particles, to be set individually for each mode
  m0 = slMass(i);
  mass(1) = slMass(j);

  switch (q) {
    case UPQUARK:
      if(k == 1) mass(2) = MUP;
      else if(k == 2) mass(2) = MCHARM;
      else if(k == 3) mass(2) = model->displayDataSet().displayPoleMt();
      else return 0;
      break;
    case DOWNQUARK:
      if(k == 1) mass(2) = MDOWN;
      else if(k == 2) mass(2) = MSTRANGE;
      else if(k == 3) mass(2) = MBOTTOM;
      else return 0;
      break;
    default:
      return 0;
  }

  mass(3) = mass(2);

  //Restrict to kinematically allowed decays
  if(m0 <= mass(1) + mass(2) + mass(3)) return 0;

  // Restrict decays to sleptons which cannot decay to any on-shell neutralino
  //(assumes lightest neutralino is lighter than light chargino)
  if(m0 >= fabs(model->displayPhys().mneut.display(1)) + MELECTRON)
    return 0;

  width = 0;

  /////////////////////////////// ZZ ////////////////////////////////////////

  mass(4) = model->displayMz();
  mass(5) = mass(4);
  cZZ(1) = ( sqr(LeftZCoupling(q)) + sqr(RightZCoupling(q)) ) 
    * sqr(Zeta2(j,i));
  cZZ(2) = 2. * LeftZCoupling(q) * RightZCoupling(q) 
    * sqr(Zeta2(j,i));
  integrals->setMasses(mass,m0);
  widthZZ = integrals->evaluateGaugeIntegrals(cZZ);

  widthZZ *= 9 * m0 / (128 * PI * PI * PI);

  /////////////////////////// HiggsHiggs //////////////////////////////////////

  widthHH = 0;

  //h0h0 contribution
  mass(4) = model->displayPhys().mhiggs.display(1);
  mass(5) = mass(4);
  cHH = sqr(Sigma2(j,i) * HiggsCoupling(SMHIGGS,q,k));
  integrals->setMasses(mass,m0);
  widthHH += integrals->evaluateScalarIntegral(cHH);

  //H0H0 contribution
  mass(4) = model->displayPhys().mhiggs.display(3);
  mass(5) = mass(4);
  cHH = sqr(Sigma3(j,i) * HiggsCoupling(HEAVYHIGGS,q,k));
  integrals->setMasses(mass,m0);
  widthHH += integrals->evaluateScalarIntegral(cHH);

  //A0A0 contribution
  mass(4) = model->displayPhys().mhiggs.display(2);
  mass(5) = mass(4);
  cHH = sqr(Sigma4(j,i) * HiggsCoupling(CPODDHIGGS,q,k));
  integrals->setMasses(mass,m0);
  widthHH += integrals->evaluatePseudoscalarIntegral(cHH);

  //G0G0 contribution
  mass(4) = model->displayMz();
  mass(5) = mass(4);
  cHH = sqr(Sigma5(j,i) * HiggsCoupling(G0HIGGS,q,k));
  integrals->setMasses(mass,m0);
  widthHH += integrals->evaluatePseudoscalarIntegral(cHH);

  // Crossed Matrix Elements for HH

  //h0H0 contribution
  mass(4) = model->displayPhys().mhiggs.display(1);
  mass(5) = model->displayPhys().mhiggs.display(3);
  cHH = Sigma2(j,i) * Sigma3(j,i) * HiggsCoupling(SMHIGGS,q,k)
    * HiggsCoupling(HEAVYHIGGS,q,k);
  integrals->setMasses(mass,m0);
  widthHH += 2 * integrals->evaluateScalarScalarCrossIntegral(cHH);

  //A0G0 contribution
  mass(4) = model->displayPhys().mhiggs.display(2);
  mass(5) = model->displayMz();
  cHH = Sigma4(j,i) * Sigma5(j,i) * HiggsCoupling(CPODDHIGGS,q,k)
    * HiggsCoupling(G0HIGGS,q,k);
  integrals->setMasses(mass,m0);
  widthHH += 2 * integrals->evaluatePseudoscalarPseudoscalarCrossIntegral(cHH);

  //Scale width by appropriate factors
  widthHH *= 9.0 / (128 * PI * PI * PI * m0);

  ///////////////////////////////// ZH //////////////////////////////////////

  widthZH = 0;

  //Zh0 contribution
  mass(4) = model->displayMz();
  mass(5) = model->displayPhys().mhiggs.display(1);
  cZH = Sigma2(j,i) * Zeta2(j,i) * HiggsCoupling(SMHIGGS,q,k)
    * ( LeftZCoupling(q) + RightZCoupling(q) );
  integrals->setMasses(mass,m0);
  widthZH += integrals->evaluateScalarGaugeCrossIntegral(cZH);

  //ZH0 contribution
  mass(4) = model->displayMz();
  mass(5) = model->displayPhys().mhiggs.display(3);
  cZH = Sigma3(j,i) * Zeta2(j,i) * HiggsCoupling(HEAVYHIGGS,q,k)
    * ( LeftZCoupling(q) + RightZCoupling(q) );
  integrals->setMasses(mass,m0);
  widthZH += integrals->evaluateScalarGaugeCrossIntegral(cZH);

  //ZA0 contribution
  mass(4) = model->displayMz();
  mass(5) = model->displayPhys().mhiggs.display(2);
  cZH = - Sigma4(j,i) * Zeta2(j,i) * HiggsCoupling(CPODDHIGGS,q,k)
    * ( LeftZCoupling(q) - RightZCoupling(q) );
  integrals->setMasses(mass,m0);
  widthZH += integrals->evaluateA0GaugeCrossIntegral(cZH);

  //ZG0 contribution
  mass(4) = model->displayMz();
  mass(5) = mass(4);
  cZH = - Sigma5(j,i) * Zeta2(j,i) * HiggsCoupling(G0HIGGS,q,k)
    * ( LeftZCoupling(q) - RightZCoupling(q) );
  integrals->setMasses(mass,m0);
  widthZH += integrals->evaluateG0GaugeCrossIntegral(cZH);

  //Scale width by appropriate factors
  widthZH *= 9.0 / (128 * PI * PI * PI);

  /////////////////////////////// Total /////////////////////////////////////

  return widthZZ + widthHH + 2 * widthZH;
}

//Calculates all slepton decays and passes them to the DecayFormat object 
//for output.  Only positive-definite decays are output; decays with widths 
//of zero are discarded.
void DecayCalc::OutputSleptonDecays() {
  double width;
  int i,j,k,m;

  for(i = 1; i <= 6; i++) {
    cout << "Calculating widths for slepton " << i << " of 6" << endl;
    for(j = 1; j <= 4; j++) {
      for(k = 1; k <= 3; k++) {
	//[5]: \tilde{l}_i -> \chi_j^0 l_k
        width = CalcSlepChi0Lep(i,j,k);
        if (width > 0) {
          decayOut->TwoBodyDecay(width,SLEPTON,i,NEUTRALINO,j,LEPTON,k);
          decayOut->TwoBodyDecay(width,ANTISLEPTON,i,NEUTRALINO,j,
                                 ANTILEPTON,k);
        }
      }
    }

    for(j = 1; j <= 2; j++) {
      for(k = 1; k <= 3; k++) {
	//[6]: \tilde{l}_i -> \chi^-_j \nu_k
        width = CalcSlepChiMinusNu(i,j,k);
        if (width > 0) {
          decayOut->TwoBodyDecay(width,SLEPTON,i,CHARGINOMINUS,j,NEUTRINO,k);
          decayOut->TwoBodyDecay(width,ANTISLEPTON,i,CHARGINOPLUS,j,
                                ANTINEUTRINO,k);
        }
      }
    }

    for(j = 1; j <= 6; j++) {
      //[7]: \tilde{l}_i -> \tilde{l}_j h^0
      width = CalcSlepSlepHiggs(i,j);
      if (width > 0) {
        decayOut->TwoBodyDecay(width,SLEPTON,i,SLEPTON,j,SMHIGGS,0);
        decayOut->TwoBodyDecay(width,ANTISLEPTON,i,ANTISLEPTON,j,SMHIGGS,0);
      }
      //[8]: \tilde{l}_i -> \tilde{l}_j Z
      width = CalcSlepSlepZ(i,j);
      if (width > 0) {
        decayOut->TwoBodyDecay(width,SLEPTON,i,SLEPTON,j,SMZ,0);
        decayOut->TwoBodyDecay(width,ANTISLEPTON,i,ANTISLEPTON,j,SMZ,0);
      }
    }

    for(j = 1; j <= 3; j++) {
      //[9]: \tilde{l}_i -> \tilde{\nu}_j W^-
      width = CalcSlepSnuWMinus(i,j);
      if (width > 0) {
        decayOut->TwoBodyDecay(width,SLEPTON,i,SNEUTRINO,j,WMINUS,0);
        decayOut->TwoBodyDecay(width,ANTISLEPTON,i,ANTISNEUTRINO,j,WPLUS,0);
      }
    }

    for(j = 1; j <= 6; j++) {
      for(k = 1; k <= 3; k++) {
        for(m = 1; m <= 3; m++) {
	  //[17]: \tilde{l}_i -> \tilde{l}_j + l_k + \overline{l}_m
          width = CalcSlepSlepLepLepBar(i,j,k,m);
          if(width > 0) {
            decayOut->ThreeBodyDecay(width,SLEPTON,i,SLEPTON,j,
                                    LEPTON,k,ANTILEPTON,m);
            decayOut->ThreeBodyDecay(width,ANTISLEPTON,i,ANTISLEPTON,j,
                                    ANTILEPTON,k,LEPTON,m);
          }
	  //[19]: \tilde{l}_i -> \tilde{l}_j + nu_k + nubar_m
	  //[18]: \tilde{l}_i -> \tilde{l}_j^* + l_k + l_m/
          width = CalcSlepSlepNuNuBar(i,j,k,m);
          if(width > 0) {
            decayOut->ThreeBodyDecay(width,SLEPTON,i,SLEPTON,j,NEUTRINO,k,
                                     ANTINEUTRINO,m);
            decayOut->ThreeBodyDecay(width,ANTISLEPTON,i,ANTISLEPTON,j,
                                     ANTINEUTRINO,k,NEUTRINO,m);
	  }
        }

        for(m = k; m <= 3; m++) {
	  //[18]: \tilde{l}_i -> \tilde{l}_j^* + l_k + l_m
          width = CalcSlepSlepBarLepLep(i,j,k,m);
          if(width > 0) {
            decayOut->ThreeBodyDecay(width,SLEPTON,i,ANTISLEPTON,j,
                                    LEPTON,k,LEPTON,m);
            decayOut->ThreeBodyDecay(width,ANTISLEPTON,i,SLEPTON,j,
                                    ANTILEPTON,k,ANTILEPTON,m);
          }
        }
	//[20]: \tilde{l}_i -> \tilde{l}_j^* + q \overline{q}
	width = CalcSlepSlepQQBar(i,j,k,UPQUARK);
	if(width > 0) {
	  decayOut->ThreeBodyDecay(width,SLEPTON,i,SLEPTON,j,
				   UPQUARK,k,ANTIUPQUARK,k);
	  decayOut->ThreeBodyDecay(width,ANTISLEPTON,i,ANTISLEPTON,j,
				   UPQUARK,k,ANTIUPQUARK,k);
	}
	width = CalcSlepSlepQQBar(i,j,k,DOWNQUARK);
	if(width > 0) {
	  decayOut->ThreeBodyDecay(width,SLEPTON,i,SLEPTON,j,
				   DOWNQUARK,k,ANTIDOWNQUARK,k);
	  decayOut->ThreeBodyDecay(width,ANTISLEPTON,i,ANTISLEPTON,j,
				   DOWNQUARK,k,ANTIDOWNQUARK,k);
	}
      }
    }
  }
}

// Evaluates [10]: \tilde{\nu}_i -> \chi_j \nu_k
double DecayCalc::CalcSnuChi0Nu(int i,int j,int k) {
  double m_i, m_j, alpha, lambda, width;

  //Assign masses
  m_i = snuMass(i);
  m_j = model->displayPhys().mneut.display(j);

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j)) return 0;

  alpha = Alpha(j,i,k);

  width = 1 / (16 * PI * pow(m_i,3)) * sqr(alpha) * sqr(sqr(m_i) - sqr(m_j));

  return width;
}

// Evaluates [11]: \tilde{\nu}_i -> \chi_j^+ l_k
double DecayCalc::CalcSnuChiPlusLep(int i,int j,int k) {
  double m_i, m_j, m_k, delta1, delta2, lambda, width;

  //Assign masses
  m_i = snuMass(i);
  m_j = model->displayPhys().mch.display(j);
  m_k = LepMass(k);

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j) + fabs(m_k)) return 0;

  lambda = Lambda(sqr(m_i), sqr(m_j), sqr(m_k));
  delta1 = Delta1(j,i,k);
  delta2 = Delta2(j,i,k);

  width = (sqr(delta1) + sqr(delta2)) * (sqr(m_i) - sqr(m_j) - sqr(m_k))
         + 4 * m_j * m_k * delta1 * delta2;
  width *= 1 / (16. * PI * pow(m_i,3)) * sqrt(lambda);

  return width;
}

// Evaluates [12]: \tilde{\nu}_i -> \tilde{l}_j W^+
double DecayCalc::CalcSnuSlepWPlus(int i,int j) {
  double m_i, m_j, m_W, width, lambda, zeta1;
  int count;

  //Assign masses
  m_i = snuMass(i);
  m_j = slMass(j);
  m_W = model->displayMw();

  //Restrict to kinematically allowed decays
  if (fabs(m_i) <= fabs(m_j) + fabs(m_W)) return 0;

  lambda = Lambda(sqr(m_i), sqr(m_j), sqr(m_W));
  zeta1 = Zeta1(i,j);

  width = sqr(zeta1) * pow(lambda,1.5) / (16. * PI * sqr(m_W) * pow(m_i,3.0));

  return width;
}

//Calculates all sneutrino decays and passes them to the DecayFormat object 
//for output.  Only positive-definite decays are output; decays with widths 
//of zero are discarded.
void DecayCalc::OutputSneutrinoDecays() {
  double width;
  int i,j,k;

  for(i = 1; i <= 3; i++) {
    for(j = 1; j <= 4; j++) {
      for(k = 1; k <= 3; k++) {
	//[10]: \tilde{\nu}_i -> \chi_j \nu_k
        width = CalcSnuChi0Nu(i,j,k);
        if (width > 0) {
          decayOut->TwoBodyDecay(width,SNEUTRINO,i,NEUTRALINO,j,NEUTRINO,k);
          decayOut->TwoBodyDecay(width,ANTISNEUTRINO,i,NEUTRALINO,j,
                                ANTINEUTRINO,k);
        }
      }
    }
    for(j = 1; j <= 2; j++) {
      for(k = 1; k <= 3; k++) {
	//[11]: \tilde{\nu}_i -> \chi_j^+ l_k
        width = CalcSnuChiPlusLep(i,j,k);
        if (width > 0) {
          decayOut->TwoBodyDecay(width,SNEUTRINO,i,CHARGINOPLUS,j,LEPTON,k);
          decayOut->TwoBodyDecay(width,ANTISNEUTRINO,i,CHARGINOMINUS,j,
                                ANTILEPTON,k);
        }
      }
    }
    for(j = 1; j <= 6; j++) {
      //[12]: \tilde{\nu}_i -> \tilde{l}_j W^+
      width = CalcSnuSlepWPlus(i,j);
      if (width > 0) {
        decayOut->TwoBodyDecay(width,SNEUTRINO,i,SLEPTON,j,WPLUS,0);
        decayOut->TwoBodyDecay(width,ANTISNEUTRINO,i,ANTISLEPTON,j,WMINUS,0);
      }
    }
  }
}

// Evaluates [13]: H^0 -> \tilde{l}_j \tilde{l}^*_k
double DecayCalc::CalcHSlepSlepBar(int j,int k) {
  double m_j, m_k, m_H, sigma3, width, lambda;

  //Assign masses
  m_j = slMass(j);
  m_k = slMass(k);
  m_H = model->displayPhys().mhiggs.display(3);

  //Restrict to kinematically allowed decays
  if(fabs(m_H) <= fabs(m_j) + fabs(m_k)) return 0;
  lambda = Lambda(sqr(m_H), sqr(m_j), sqr(m_k));

  sigma3 = Sigma3(j,k);

  width = sqr(sigma3) / (16 * PI * pow(m_H,3)) * sqrt(lambda);

  return width;
}

// Evaluates [14]: H^0 -> \tilde{\nu}_j \tilde{\nu}^*_k
double DecayCalc::CalcHSnuSnuBar(int j,int k) {
  double m_j, m_k, m_H, sigma1, width, lambda;

  //Assign masses
  m_j = snuMass(j);
  m_k = snuMass(k);
  m_H = model->displayPhys().mhiggs.display(3);

  //Restrict to kinematically allowed decays
  if(fabs(m_H) <= fabs(m_j) + fabs(m_k)) return 0;

  lambda = Lambda(sqr(m_H), sqr(m_j), sqr(m_k));
  sigma1 = Sigma1(j,k);

  width = sqr(sigma1) / (16 * PI * pow(m_H,3)) * sqrt(lambda);
  return width;
}

// Evaluates [15]: A^0 -> \tilde{l}_j \tilde{l}^*_k
double DecayCalc::CalcASlepSlepBar(int j,int k) {
  double m_j, m_k, m_A, sigma4, width, lambda;

  //Assign masses
  m_j = slMass(j);
  m_k = slMass(k);
  m_A = model->displayPhys().mhiggs.display(2);

  //Restrict to kinematically allowed decays
  if(fabs(m_A) <= fabs(m_j) + fabs(m_k)) return 0;

  lambda = Lambda(sqr(m_A), sqr(m_j), sqr(m_k));
  sigma4 = Sigma4(j,k);

  width = sqr(sigma4) / (16 * PI * pow(m_A,3)) * sqrt(lambda);

  return width;
}

// Evaluates [16]: H^- -> \tilde{l}_j \tilde{\nu}_k
double DecayCalc::CalcHMinusSlepSnuBar(int j,int k) {
  double m_j, m_k, m_HM, rho, width, lambda;

  //Assign masses
  m_j = slMass(j);
  m_k = snuMass(k);
  m_HM = model->displayPhys().mhiggs.display(4);

  //Restrict to kinematically allowed decays
  if(fabs(m_HM) <= fabs(m_j) + fabs(m_k)) return 0;

  lambda = Lambda(sqr(m_HM), sqr(m_j), sqr(m_k));
  rho = Rho(k,j);

  width = sqr(rho) / (16 * PI * pow(m_HM,3)) * sqrt(lambda);

  return width;
}

//Calculates all Higgs decays and passes them to the DecayFormat object 
//for output.  Only positive-definite decays are output; decays with widths 
//of zero are discarded.  We assume the light real Higgs scalar is too light
//too decay into slepton/sneutrino pairs.
void DecayCalc::OutputHiggsDecays() {
  double width;
  int j,k;

  for(j = 1; j <= 6; j++) {
    for(k = 1; k <= 6; k++) {
      //[13]: H^0 -> \tilde{l}_j \tilde{l}^*_k
      width = CalcHSlepSlepBar(j,k);
      if (width > 0) {
        decayOut->TwoBodyDecay(width,HEAVYHIGGS,0,SLEPTON,j,ANTISLEPTON,k);
      }
    }
  }
  for(j = 1; j <= 3; j++) {
    for(k = 1; k <= 3; k++) {
      //[14]: H^0 -> \tilde{\nu}_j \tilde{\nu}^*_k
      width = CalcHSnuSnuBar(j,k);
      if (width > 0) {
        decayOut->TwoBodyDecay(width,HEAVYHIGGS,0,SNEUTRINO,j,ANTISNEUTRINO,k);
      }
    }
  }
  for(j = 1; j <= 6; j++) {
    for(k = 1; k <= 6; k++) {
      //[15]: A^0 -> \tilde{l}_j \tilde{l}^*_k
      width = CalcASlepSlepBar(j,k);
      if (width > 0) {
        decayOut->TwoBodyDecay(width,CPODDHIGGS,0,SLEPTON,j,ANTISLEPTON,k);
      }
    }
    for(k = 1; k <= 3; k++) {
      //[16]: H^- -> \tilde{l}_j \tilde{\nu}_k
      width = CalcHMinusSlepSnuBar(j,k);
      if (width > 0) {
        decayOut->TwoBodyDecay(width,HIGGSMINUS,0,SLEPTON,j,ANTISNEUTRINO,k);
        decayOut->TwoBodyDecay(width,HIGGSPLUS,0,ANTISLEPTON,j,SNEUTRINO,k);
      }
    }
  }
}

//Returns the mass of the kth generation lepton
double DecayCalc::LepMass(int k) {
  switch(k) {
    case 1: return MELECTRON;
    case 2: return MMUON;
    case 3: return MTAU;
    default: return 0;
  }
}

//Common phase space function lambda
double DecayCalc::Lambda(double x, double y, double z) {
  double lambda = (sqr(x) + sqr(y) + sqr(z) - 2*x*y - 2*x*z - 2*y*z);

  //Correct for negative lambda values
  if(lambda < 0) {
    //Return an error message if the negative lambda value is large
    //If it is small, the value is assumed to be zero but found to be slightly
    //negative due to machine precision.
    if(fabs(lambda) > 1.0e-12)
      cout << "Warning: lambda value less than 0 found: " << lambda << endl;
    return 0;
  }

  return lambda;
}

// Calculate the parameter alpha for given particle eigenstates
double DecayCalc::Alpha(int chi, int snu, int nu) {
  double g, gp, alpha;

  gp = model->displayGaugeCoupling(1)*sqrt(3.0/5.0);
  g = model->displayGaugeCoupling(2);

  alpha = - 1./root2 * snuRot(nu,snu) 
    * (g * model->displayPhys().mixNeut.display(2,chi)
       - gp * model->displayPhys().mixNeut.display(1,chi));

  return alpha;
}

// Calculate the parameter beta1 for given particle eigenstates
double DecayCalc::Beta1(int chi, int slep, int lep) {
  double g, gp, beta1, yukawa;

  g = model->displayGaugeCoupling(2);
  gp = model->displayGaugeCoupling(1)*sqrt(3.0/5.0);
  yukawa = model->displayYukawaElement(YE,lep,lep);

  beta1 = 1./root2 * slRot(lep,slep) 
    * (g * model->displayPhys().mixNeut.display(2,chi)
       + gp * model->displayPhys().mixNeut.display(1,chi))
      - yukawa * model->displayPhys().mixNeut.display(3,chi) 
      * slRot(lep+3,slep);

  return beta1;
}

// Calculate the parameter beta2 for given particle eigenstates
double DecayCalc::Beta2(int chi, int slep, int lep) {
  double gp, beta2, yukawa;

  gp = model->displayGaugeCoupling(1)*sqrt(3.0/5.0);
  yukawa = model->displayYukawaElement(YE,lep,lep);

  beta2 = - root2 * gp * slRot(lep+3,slep) 
      * model->displayPhys().mixNeut.display(1,chi)
      - yukawa * model->displayPhys().mixNeut.display(3,chi) * slRot(lep,slep);

  return beta2;
}

// Calculate the parameter gamma for given particle eigenstates
double DecayCalc::Gamma(int chim, int slep, int nu) {
  double gamma, thetaL, g, yukawa;

  thetaL = model->displayPhys().thetaL;

  g = model->displayGaugeCoupling(2);
  yukawa = model->displayYukawaElement(YE,nu,nu);

  if(chim == 1) {
    gamma = - g * cos(thetaL) * slRot(nu,slep)
            + sin(thetaL) * yukawa * slRot(nu + 3,slep);
  }
  else if(chim == 2) {
    gamma = g * sin(thetaL) * slRot(nu,slep)
            + cos(thetaL) * yukawa * slRot(nu + 3,slep);
  }
  else {
    cout << "Calling gamma with chargino index out of bounds" << endl;
    return 0;
  }

  return gamma;
}

// Calculate the parameter delta1 for given particle eigenstates
double DecayCalc::Delta1(int chip, int snu, int lep) {
  double delta1, thetaR, g;

  g = model->displayGaugeCoupling(2);
  thetaR = model->displayPhys().thetaR;

  if(chip == 1) {
    delta1 = - g * cos(thetaR) * snuRot(lep,snu);
  }
  else if(chip == 2) {
    delta1 = g * sin(thetaR) * snuRot(lep,snu);
  }
  else {
    cout << "Calling delta1 with chargino index out of bounds" << endl;
    return 0;
  }

  return delta1;
}

// Calculate the parameter delta2 for given particle eigenstates
double DecayCalc::Delta2(int chip, int snu, int lep) {
  double delta2, thetaL, yukawa;

  thetaL = model->displayPhys().thetaL;
  yukawa = model->displayYukawaElement(YE,lep,lep);

  if(chip == 1) {
    delta2 = sin(thetaL) * yukawa * snuRot(lep,snu);
  }
  else if (chip == 2) {
    delta2 = cos(thetaL) * yukawa * snuRot(lep,snu);
  }
  else {
    cout << "Calling delta2 with chargino index out of bounds" << endl;
    return 0;
  }

  return delta2;
}

// Calculate the parameter sigma1 for given particle eigenstates
double DecayCalc::Sigma1(int snu1, int snu2) {
  double sigma1, M_w, M_z, g, costhetaw, thetaH, beta;
  DoubleVector m(3);

  M_w = model->displayMw();
  M_z = model->displayMz();
  g = model->displayGaugeCoupling(2);
  costhetaw = M_w/M_z;
  beta = atan(model->displayTanb());
  thetaH = model->displayPhys().thetaH;

  if(snu1 == snu2)
    sigma1 = - g * M_w / (2 * sqr(costhetaw)) * cos(beta + thetaH);
  else
    sigma1 = 0;

  return sigma1;
}

// Calculate the parameter sigma2 for given particle eigenstates
double DecayCalc::Sigma2(int slep1, int slep2) {
  double sigma2, M_w, M_z, g, sin2thetaw, tanthetaw, thetaH, beta, mu;
  int count;
  DoubleVector m(3),A(3);

  m(1) = MELECTRON;
  m(2) = MMUON;
  m(3) = MTAU;

  A(1) = model->displaySoftA(EA,1,1);
  A(2) = model->displaySoftA(EA,2,2);
  A(3) = model->displaySoftA(EA,3,3);

  M_w = model->displayMw();
  M_z = model->displayMz();
  g = model->displayGaugeCoupling(2);
  sin2thetaw = 1 - sqr(M_w/M_z);
  tanthetaw = sqrt(sin2thetaw / (1 - sin2thetaw));
  beta = atan(model->displayTanb());
  thetaH = model->displayPhys().thetaH;
  mu = model->displaySusyMu();

  sigma2 = 0;

  //Sum over the intermediate lepton index
  for(count = 1; count <= 3; count++) {
    sigma2 += - g * slRot(count,slep1) * slRot(count,slep2)
      * (M_w/2 * (1 - sqr(tanthetaw)) * sin(thetaH+beta)
         - sqr(m(count))*sin(thetaH) / (M_w*cos(beta)));
    sigma2 += - g * slRot(count+3,slep1) * slRot(count+3,slep2)
      * (M_w * sqr(tanthetaw) * sin(thetaH+beta)
         - sqr(m(count))*sin(thetaH) / (M_w*cos(beta)));
    sigma2 += g * m(count) / (2*M_w*cos(beta)) 
      * (slRot(count,slep1) * slRot(count+3,slep2) + 
         slRot(count+3,slep1) * slRot(count,slep2))
      * (mu*cos(thetaH) + A(count)*sin(thetaH));
  }

  return sigma2;
}

// Calculate the parameter sigma3 for given particle eigenstates
double DecayCalc::Sigma3(int slep1, int slep2) {
  double sigma3, M_w, M_z, g, sin2thetaw, tanthetaw, thetaH, beta, mu;
  int count;
  DoubleVector m(3),A(3);

  m(1) = MELECTRON;
  m(2) = MMUON;
  m(3) = MTAU;

  A(1) = model->displaySoftA(EA,1,1);
  A(2) = model->displaySoftA(EA,2,2);
  A(3) = model->displaySoftA(EA,3,3);

  M_w = model->displayMw();
  M_z = model->displayMz();
  g = model->displayGaugeCoupling(2);
  sin2thetaw = 1 - sqr(M_w/M_z);
  tanthetaw = sqrt(sin2thetaw / (1 - sin2thetaw));
  beta = atan(model->displayTanb());
  thetaH = model->displayPhys().thetaH;
  mu = model->displaySusyMu();

  sigma3 = 0;

  //Sum over the intermediate lepton index
  for(count = 1; count <= 3; count++) {
    sigma3 += g * slRot(count,slep1) * slRot(count,slep2)
      * (M_w/2 * (1 - sqr(tanthetaw)) * cos(thetaH+beta)
         - sqr(m(count))*cos(thetaH) / (M_w*cos(beta)));
    sigma3 += g * slRot(count+3,slep1) * slRot(count+3,slep2)
      * (M_w * sqr(tanthetaw) * cos(thetaH+beta)
         - sqr(m(count))*cos(thetaH) / (M_w*cos(beta)));
    sigma3 += g * m(count) / (2*M_w*cos(beta)) 
      * (slRot(count,slep1) * slRot(count+3,slep2) + 
         slRot(count+3,slep1) * slRot(count,slep2))
         * (mu*sin(thetaH) - A(count)*cos(thetaH));
  }

  return sigma3;
}

// Calculate the parameter sigma4 for given particle eigenstates
double DecayCalc::Sigma4(int slep1, int slep2) {
  double sigma4, M_w, g, beta, mu;
  int count;
  DoubleVector m(3),A(3);

  m(1) = MELECTRON;
  m(2) = MMUON;
  m(3) = MTAU;

  A(1) = model->displaySoftA(EA,1,1);
  A(2) = model->displaySoftA(EA,2,2);
  A(3) = model->displaySoftA(EA,3,3);

  M_w = model->displayMw();
  g = model->displayGaugeCoupling(2);
  beta = atan(model->displayTanb());
  mu = model->displaySusyMu(); 

  sigma4 = 0;
  //Sum over the intermediate lepton index

  for(count = 1; count <= 3; count++) {
    sigma4 += g * m(count)/(2*M_w) * (mu + A(count)*tan(beta))
      * (slRot(count,slep1) * slRot(count+3,slep2) - 
         slRot(count+3,slep1) * slRot(count,slep2));
  }

  return sigma4;
}

// Calculate the parameter sigma5 for given particle eigenstates
double DecayCalc::Sigma5(int slep1, int slep2) {
  double sigma5, M_w, g, beta, mu;
  int count;
  DoubleVector m(3),A(3);

  m(1) = MELECTRON;
  m(2) = MMUON;
  m(3) = MTAU;

  A(1) = model->displaySoftA(EA,1,1);
  A(2) = model->displaySoftA(EA,2,2);
  A(3) = model->displaySoftA(EA,3,3);

  M_w = model->displayMw();
  g = model->displayGaugeCoupling(2);
  beta = atan(model->displayTanb());
  mu = model->displaySusyMu(); 

  sigma5 = 0;

  for(count = 1; count <= 3; count++) {
    sigma5 += g * m(count)/(2*M_w) * (mu * tan(beta) + A(count))
      * (slRot(count,slep1) * slRot(count+3,slep2) - 
         slRot(count+3,slep1) * slRot(count,slep2));
  }

  return sigma5;
}

// Calculate the parameter rho for given particle eigenstates
double DecayCalc::Rho(int snu, int slep) {
  double rho, M_w, g, mu, beta;
  int count;
  DoubleVector m(3),A(3);

  m(1) = MELECTRON;
  m(2) = MMUON;
  m(3) = MTAU;

  A(1) = model->displaySoftA(EA,1,1);
  A(2) = model->displaySoftA(EA,2,2);
  A(3) = model->displaySoftA(EA,3,3);

  M_w = model->displayMw();
  g = model->displayGaugeCoupling(2);
  beta = atan(model->displayTanb());
  mu = model->displaySusyMu();

  rho = 0;

  //Sum over the intermediate lepton index
  for(count = 1; count <= 3; count++) {
    rho += - g / root2 * snuRot(count,snu) * slRot(count,slep)
      * (M_w * sin(2*beta) - sqr(m(count))*tan(beta)/M_w);
    rho += g * m(count)/root2/M_w * snuRot(count,snu) * slRot(count+3,slep) 
      * (mu + A(count) * tan(beta));
  }

  return rho;
}

// Calculate the parameter zeta1 for given particle eigenstates
double DecayCalc::Zeta1(int snu, int slep) {
  double g, zeta1;
  int count;

  g = model->displayGaugeCoupling(2);

  zeta1 = 0;

  //Sum over the intermediate lepton index
  for(count = 1; count <= 3; count++) {
    zeta1 += snuRot(count,snu) * slRot(count,slep);
  }

  zeta1 *= - g / root2;

  return zeta1;
}

// Calculate the parameter zeta2 for given particle eigenstates
double DecayCalc::Zeta2(int slep1, int slep2) {
  double gp, g, tanthetaW, sinthetaW, costhetaW, zeta2;
  int count;

  gp = model->displayGaugeCoupling(1) * sqrt(3.0/5.0);
  g = model->displayGaugeCoupling(2);
  tanthetaW = gp/g;
  sinthetaW = gp / sqrt(g * g + gp * gp);
  costhetaW = g / sqrt(g * g + gp * gp);

  zeta2 = 0;

  //Sum over the intermediate lepton index
  for(count = 1; count <= 3; count++) {
    zeta2 += slRot(count,slep1) * slRot(count,slep2);
  }

  if(slep1 == slep2)
    zeta2 -= 2 * sqr(sinthetaW);

  zeta2 *= g / (2 * costhetaW);

  return zeta2; 
}

//Calculate the coupling of the Z to a left-handed fermion of the given type
double DecayCalc::LeftZCoupling(particletype daughter) {
  double gp, g, sinThetaW, cosThetaW, T3, Q;

  switch (daughter){
    case NEUTRINO: // neutrinos
      T3 = 0.5;
      Q = 0.;
      break;
    case LEPTON: // leptons
      T3 = - 0.5;
      Q = - 1.;
      break;
    case UPQUARK: // up type quarks
      T3 = 0.5;
      Q = 2./3.;
      break;
    case DOWNQUARK: // down type quarks
      T3 = -0.5;
      Q = -1./3.;
      break;
    default: return 0;
  }

  gp = model->displayGaugeCoupling(1) * sqrt(3.0/5.0);
  g = model->displayGaugeCoupling(2);
  cosThetaW = g / sqrt( sqr(g) + sqr(gp) );
  sinThetaW = gp / sqrt( sqr(g) + sqr(gp) );

  return g * (T3 - Q * sqr(sinThetaW)) / cosThetaW;
}

//Calculate the coupling of the Z to a right-handed fermion of the given type
double DecayCalc::RightZCoupling(particletype daughter) {
  double gp, g, sinThetaW, cosThetaW, Q;

  switch (daughter){
    case LEPTON: // leptons
      Q = - 1.;
      break;
    case UPQUARK: // up type quarks
       Q = 2./3.;
      break;
    case DOWNQUARK: // down type quarks
       Q = -1./3.;
      break;
    default: return 0;
  }

  gp = model->displayGaugeCoupling(1) * sqrt(3.0/5.0);
  g = model->displayGaugeCoupling(2);
  cosThetaW = g / sqrt( sqr(g) + sqr(gp) );
  sinThetaW = gp / sqrt( sqr(g) + sqr(gp) );

  return - g * Q * sqr(sinThetaW) / cosThetaW;
}

//Calculate the coupling of the given Higgs to a fermion of the given type
double DecayCalc::HiggsCoupling(particletype parent, particletype daughter,
				int gen) {
  double yuk, thetaH, beta, trigfactor;
  thetaH = model->displayPhys().thetaH;
  beta = atan(model->displayTanb());

  switch(daughter) {
    case NEUTRINO: return 0;
    case LEPTON: 
      yuk = model->displayYukawaElement(YE,gen,gen);
      if(parent == SMHIGGS) trigfactor = - sin(thetaH);
      else if(parent == HEAVYHIGGS) trigfactor = cos(thetaH);
      else if(parent == CPODDHIGGS) trigfactor = - sin(beta);
      else if(parent == G0HIGGS) trigfactor = cos(beta);
      else return 0;
      break;
    case UPQUARK:
      yuk = model->displayYukawaElement(YU,gen,gen);
      if(parent == SMHIGGS) trigfactor = cos(thetaH);
      else if(parent == HEAVYHIGGS) trigfactor = sin(thetaH);
      else if(parent == CPODDHIGGS) trigfactor = cos(beta);
      else if(parent == G0HIGGS) trigfactor = sin(beta);
      else return 0;
      break;
    case DOWNQUARK:
      yuk = model->displayYukawaElement(YD,gen,gen);
      if(parent == SMHIGGS) trigfactor = - sin(thetaH);
      else if(parent == HEAVYHIGGS) trigfactor = cos(thetaH);
      else if(parent == CPODDHIGGS) trigfactor = - sin(beta);
      else if(parent == G0HIGGS) trigfactor = cos(beta);
      else return 0;
      break;
    default: return 0;
  }

  return yuk * trigfactor / root2;
}

//Helper method for the three-body decays.  Produces a product of the form
// \beta_{aik}^{n1} \beta_{ajm}^{n2} \beta_{bik}^{n3} \beta_{bjm}^{n4}
double DecayCalc::IntegralCoefficient(int i, int j, int k, int m,int a,int b,
                                      int n1, int n2, int n3, int n4) {

  double coeff1 = 1., coeff2 = 1.;

  if(n1 == 1) {
    coeff1 *= Beta1(a,i,k);
    coeff2 *= Beta2(a,i,k);
  }
  else {
    coeff1 *= Beta2(a,i,k);
    coeff2 *= Beta1(a,i,k);
  }

  if(n2 == 1) {
    coeff1 *= Beta1(a,j,m);
    coeff2 *= Beta2(a,j,m);
  }
  else {
    coeff1 *= Beta2(a,j,m);
    coeff2 *= Beta1(a,j,m);
  }

  if(n3 == 1) {
    coeff1 *= Beta1(b,i,k);
    coeff2 *= Beta2(b,i,k);
  }
  else {
    coeff1 *= Beta2(b,i,k);
    coeff2 *= Beta1(b,i,k);
  }

  if(n4 == 1) {
    coeff1 *= Beta1(b,j,m);
    coeff2 *= Beta2(b,j,m);
  }
  else {
    coeff1 *= Beta2(b,j,m);
    coeff2 *= Beta1(b,j,m);
  }

  return coeff1 + coeff2;
}

//Helper method for the three-body decays.  Produces a product of the form
// \beta_{aim}^{n1} \beta_{ajk}^{n2} \beta_{bik}^{n3} \beta_{bjm}^{n4}
double DecayCalc::CrossCoefficient(int i, int j, int k, int m, int a, int b, 
                                   int n1, int n2, int n3, int n4) {

  double coeff1 = 1., coeff2 = 1.;

  if(n1 == 1) {
    coeff1 *= Beta1(a,i,m);
    coeff2 *= Beta2(a,i,m);
  }
  else {
    coeff1 *= Beta2(a,i,m);
    coeff2 *= Beta1(a,i,m);
  }
  if(n2 == 1) {
    coeff1 *= Beta1(a,j,k);
    coeff2 *= Beta2(a,j,k);
  }
  else {
    coeff1 *= Beta2(a,j,k);
    coeff2 *= Beta1(a,j,k);
  }
  if(n3 == 1) {
    coeff1 *= Beta1(b,i,k);
    coeff2 *= Beta2(b,i,k);
  }
  else {
    coeff1 *= Beta2(b,i,k);
    coeff2 *= Beta1(b,i,k);
  }
  if(n4 == 1) {
    coeff1 *= Beta1(b,j,m);
    coeff2 *= Beta2(b,j,m);
  }
  else {
    coeff1 *= Beta2(b,j,m);
    coeff2 *= Beta1(b,j,m);
  }

  return coeff1 + coeff2;
}
