#ifndef DECAYCALC_H
#define DECAYCALC_H
#include <flavoursoft.h>
#include <linalg.h>
#include <fstream>
#include <ParticleCodes.h>

class DecayFormat;
class ThreeBodyIntegrals;

// This is the class responsible for storing and calculating the decay rates
// involving leptons or sleptons.
class DecayCalc{
 private:
  FlavourMssmSoftsusy* model;    //Pointer to the FlavourMssmSoftsusy object 
                                 //for the model being output
  DecayFormat* decayOut;         //Pointer to the decayOut object responsible
                                 //for output of the FV slepton decays
  DoubleVector slMass,snuMass;   //Slepton and sneutrino mass vectors
  DoubleMatrix slRot,snuRot;     //Slepton and sneutrino rotation matrices

  ThreeBodyIntegrals* integrals; //Pointer to the object used for calculating
                                 //the phase space integrals for the 
                                 //three-body decays

 public:
  //Constructor - assigns pointers and diagonalizes the mass matrices
  DecayCalc(DecayFormat *output);

  //Methods for accessing class variables
  DecayFormat* getDecayOut() {return decayOut;}
  DoubleMatrix displaySlRot() const {return slRot;}
  DoubleMatrix displaySnuRot() const {return snuRot;}
  DoubleVector displaySlMass() const {return slMass;}
  DoubleVector displaySnuMass() const {return snuMass;}

  //Methods used to loop over decays of a certain type of particle and
  //output non-zero widths.
  void OutputNeutralinoDecays();
  void OutputCharginoDecays();
  void OutputSleptonDecays();
  void OutputSneutrinoDecays();
  void OutputHiggsDecays();

 private:

  //Calculate widths of neutralino decays
  double CalcChi0SlepLepBar(int i,int j,int k); // [1]
  double CalcChi0SnuNuBar(int i,int j,int k);   // [2]

  //Calculate widths of chargino decays
  double CalcChiMinusSlepNuBar(int i,int j,int k); // [3]
  double CalcChiPlusSnuLepBar(int i,int j,int k);  // [4]

  //Calculate widths of slepton decays
  double CalcSlepChi0Lep(int i,int j,int k);      // [5]
  double CalcSlepChiMinusNu(int  i,int j,int k); // [6]
  double CalcSlepSlepHiggs(int i,int j);         // [7]
  double CalcSlepSlepZ(int i,int j);             // [8]
  double CalcSlepSnuWMinus(int i,int j);        // [9]
  double CalcSlepSlepLepLepBar(int i,int j,int k,int m); // [17]
  double CalcSlepSlepBarLepLep(int i,int j,int k,int m); // [18]
  double CalcSlepSlepNuNuBar(int i,int j,int k,int m); //[19]
  double CalcSlepSlepQQBar(int i,int j,int k,particletype q); //[20]

  //Calculate widths of sneutrino decays
  double CalcSnuChi0Nu(int i,int j,int k);      // [10]
  double CalcSnuChiPlusLep(int i,int j,int k); // [11]
  double CalcSnuSlepWPlus(int i,int j);        // [12]

  //Calculate widths of Higgs decays
  double CalcHSlepSlepBar(int j,int k);     // [13]
  double CalcHSnuSnuBar(int j,int k);       // [14]
  double CalcASlepSlepBar(int j,int k);     // [15]
  double CalcHMinusSlepSnuBar(int j,int k); // [16]

  double LepMass(int k);  //Get the mass of the lepton K
  double Lambda(double x, double y, double z);  //Common phase space factor
                                                //(see manual for details)

  //Various Lagrangian coefficients (see manual for details)
  //All dependence on the slepton and sneutrino mixing matrices
  double Alpha(int chi, int snu, int nu);
  double Beta1(int chi, int slep, int lep);
  double Beta2(int chi, int slep, int lep);
  double Gamma(int chim, int slep, int lep);
  double Delta1(int chip, int slep, int lep);
  double Delta2(int chip, int slep, int lep);
  double Sigma1(int snu1, int snu2);
  double Sigma2(int slep1, int slep2);
  double Sigma3(int slep1, int slep2);
  double Sigma4(int slep1, int slep2);
  double Sigma5(int slep1, int slep2);
  double Rho(int snu, int slep);
  double Zeta1(int snu, int slep);
  double Zeta2(int slep1, int slep2);

  //Coupling of Higgs and Z bosons to Standard Model fermions
  double LeftZCoupling(particletype daughter);
  double RightZCoupling(particletype daughter);
  double HiggsCoupling(particletype parent, particletype daughter, int gen);

  //Methods to simplify the calculation of the overall multiplicative 
  //factors for the integrals in the neutralino portion of the three-body
  //decays.
  double IntegralCoefficient(int i, int j, int k, int m, int a, int b, 
                             int n1, int n2, int n3, int n4);
  double CrossCoefficient(int i, int j, int k, int m, int a, int b, 
                          int n1, int n2, int n3, int n4);
};

#endif
