///////////////////////////////////////////////////////////////////////////////
//                        DecayFormat Class
//    This class handles the output of an intermediate file that contains
//  the slepton flavor-violating decays, which will later be merged with 
//  the Susyhit output.  This class interfaces directly with the main 
//  program, and in turn contains a copy of the DecayCalc class to calculate 
//  the relevant slepton flavor-violating decay widths.  The output file 
//  produced by DecayFormat is in faux-herwig format, though it uses PDG 
//  particle codes and only contains decays which are affected by slepton 
//  flavor mixing.
///////////////////////////////////////////////////////////////////////////////

#include <DecayCalc.h>
#include <DecayFormat.h>
#include <flavoursoft.h>
#include <linalg.h>
#include <fstream>
#include <iomanip>

//Constructor for the DecayFormat class
//Stores a pointer to the passed FlavourMssmSoftsusy object, instantiates 
//a copy of the decay calculation class, and opens the file to which masses, 
//decay widths, and branching fractions will be outputted
DecayFormat::DecayFormat(FlavourMssmSoftsusy *r) {
  model = r;
  fout.open("spicydecays.out");

  decays = new DecayCalc(this);
}

//Translates from a particle type and index into the corresponding PDG
//code for the particle.  Note that for the sleptons and sneutrinos
//the mass eigenstate is used.
//The definitions of 'particletype' and the PDG codes are found in the 
//file 'ParticleCodes.h'
int DecayFormat::PDGCode(particletype particle,int particleindex) {
  int i,index;
  double max;
  switch (particle) {
  case UPQUARK:
    switch (particleindex) {
    case 1: return PDGUP;
    case 2: return PDGCHARM;
    case 3: return PDGTOP;
    default: return 0;
    }
    break;
  case ANTIUPQUARK:
    switch (particleindex) {
    case 1: return -PDGUP;
    case 2: return -PDGCHARM;
    case 3: return -PDGTOP;
    default: return 0;
    }
    break;

  case DOWNQUARK:
    switch (particleindex) {
    case 1: return PDGDOWN;
    case 2: return PDGSTRANGE;
    case 3: return PDGBOTTOM ;
    default: return 0;
    }
    break;
  case ANTIDOWNQUARK:
    switch (particleindex) {
    case 1: return -PDGDOWN;
    case 2: return -PDGSTRANGE;
    case 3: return -PDGBOTTOM ;
    default: return 0;
    }
    break;
  case LEPTON:
    switch (particleindex) {
    case 1: return PDGELECTRON;
    case 2: return PDGMUON;
    case 3: return PDGTAU;
    default: return 0;
    }
    break;
  case ANTILEPTON:
    switch (particleindex) {
    case 1: return -PDGELECTRON;
    case 2: return -PDGMUON;
    case 3: return -PDGTAU;
    default: return 0;
    }
    break;
  case NEUTRINO:
    switch (particleindex) {
    case 1: return PDGENEUTRINO;
    case 2: return PDGMUNEUTRINO;
    case 3: return PDGTAUNEUTRINO;
    default: return 0;
    }
    break;
  case ANTINEUTRINO:
    switch (particleindex) {
    case 1: return -PDGENEUTRINO;
    case 2: return -PDGMUNEUTRINO;
    case 3: return -PDGTAUNEUTRINO;
    default: return 0;
    }
    break;
  case NEUTRALINO:
    switch (particleindex) {
    case 1: return PDGNEUTR1;
    case 2: return PDGNEUTR2;
    case 3: return PDGNEUTR3;
    case 4: return PDGNEUTR4;
    default: return 0;
    }
    break;
  case CHARGINOPLUS:
    switch (particleindex) {
    case 1: return PDGCHARGINO1;
    case 2: return PDGCHARGINO2;
    default: return 0;
    }
    break;
  case CHARGINOMINUS:
    switch (particleindex) {
    case 1: return -PDGCHARGINO1;
    case 2: return -PDGCHARGINO2;
    default: return 0;
    }
    break;
  case SMHIGGS:
    return PDGHIGGS;
    break;
  case SMZ:
    return PDGZ;
    break;
  case PHOTON:
    return PDGPHOTON;
    break;
  case GLUON:
    return PDGGLUON;
    break;
  case WPLUS:
    return PDGW;
    break;
  case WMINUS:
    return -PDGW;
    break;
  case SLEPTON:
    switch (particleindex) {
    case 1: return PDGSEL;
    case 2: return PDGSMUL;
    case 3: return PDGSTAU1;
    case 4: return PDGSER;
    case 5: return PDGSMUR;
    case 6: return PDGSTAU2;
    default: return 0;
    }
    break;
  case ANTISLEPTON:
    switch (particleindex) {
    case 1: return -PDGSEL;
    case 2: return -PDGSMUL;
    case 3: return -PDGSTAU1;
    case 4: return -PDGSER;
    case 5: return -PDGSMUR;
    case 6: return -PDGSTAU2;
    default: return 0;
    }
    break;
  case SNEUTRINO:
    switch (particleindex) {
    case 1: return PDGSNUEL;
    case 2: return PDGSNUMUL;
    case 3: return PDGSNUTAUL;
    default: return 0;
    }
    break;
  case ANTISNEUTRINO:
    switch (particleindex) {
    case 1: return -PDGSNUEL;
    case 2: return -PDGSNUMUL;
    case 3: return -PDGSNUTAUL;
    default: return 0;
    }
    break;
  case HEAVYHIGGS:
    return PDGHEAVYHIGGS;
    break;
  case CPODDHIGGS:
    return PDGCPODDHIGGS;
    break;
  case HIGGSPLUS:
    return PDGCHARGEDHIGGS;
    break;
  case HIGGSMINUS:
    return -PDGCHARGEDHIGGS;
    break;
  case GRAVITON:
    return PDGGRAVITON;
    break;
  case GRAVITINO:
    return PDGGRAVITINO;
    break;
  default:
    break;
  }
}

//Output a two body decay, given the branching ratio and relevant particles.  
//Calculation and normalization must be performed before this step.
void DecayFormat::TwoBodyDecay(double width, particletype parent,
                               int parentindex, particletype daughter1,
                               int daughter1index, particletype daughter2,
                               int daughter2index) {

  int parentId, daughter1Id, daughter2Id;

  parentId = PDGCode(parent, parentindex);
  daughter1Id = PDGCode(daughter1, daughter1index);
  daughter2Id = PDGCode(daughter2, daughter2index);
  fout << setprecision(12) << "D " <<  parentId << "\t" << width << "\t" 
       << 2 << "\t" << daughter1Id << "\t" << daughter2Id << endl;
}

//Output a three body decay, given the branching ratio and relevant particles.
//Calculation and normalization must be performed before this step.
void DecayFormat::ThreeBodyDecay(double width, particletype parent,
                                 int parentindex, particletype daughter1,
                                 int daughter1index, particletype daughter2,
                                 int daughter2index, particletype daughter3,
                                 int daughter3index) {

  int parentId, daughter1Id, daughter2Id, daughter3Id;

  parentId = PDGCode(parent, parentindex);
  daughter1Id = PDGCode(daughter1, daughter1index);
  daughter2Id = PDGCode(daughter2, daughter2index);
  daughter3Id = PDGCode(daughter3, daughter3index);
  fout << setprecision(12) << "D " << parentId << "\t" << width << "\t" 
       << 3 << "\t" << daughter1Id << "\t" << daughter2Id << "\t" 
       << daughter3Id << endl;
}

// Outputs a particle code, mass and decay rate to the file.
void DecayFormat::OutputParticle(int particleId,double mass) {
  fout << setprecision(12) << particleId << " " << mass << endl;
}

// Output the herwig particle data. For each particle the mass and a 
// a zero decay rate are output.  Note that for the sleptons we output 
// mass eigenstates.
void DecayFormat::OutputParticleData() {
  cout << "Output Particle Data: Start" << endl;
  fout << setprecision(12);
  fout << "65" << endl;
  OutputParticle(PDGHIGGS,model->displayPhys().mhiggs.display(1));
  OutputParticle(PDGHEAVYHIGGS,model->displayPhys().mhiggs.display(3));
  OutputParticle(PDGCPODDHIGGS,model->displayPhys().mhiggs.display(2));
  OutputParticle(PDGCHARGEDHIGGS,model->displayPhys().mhiggs.display(4));
  OutputParticle(-PDGCHARGEDHIGGS,model->displayPhys().mhiggs.display(4));
  OutputParticle(PDGSDL,model->displayPhys().md.display(1,1));
  OutputParticle(PDGSUL,model->displayPhys().mu.display(1,1));
  OutputParticle(PDGSSL,model->displayPhys().md.display(1,2));
  OutputParticle(PDGSCL,model->displayPhys().mu.display(1,2));
  OutputParticle(PDGSB1,min(model->displayPhys().md.display(1,3),
                               model->displayPhys().md.display(2,3)));
  OutputParticle(PDGST1,min(model->displayPhys().mu.display(1,3),
                               model->displayPhys().mu.display(2,3)));
  OutputParticle(-PDGSDL,model->displayPhys().md.display(1,1));
  OutputParticle(-PDGSUL,model->displayPhys().mu.display(1,1));
  OutputParticle(-PDGSSL,model->displayPhys().md.display(1,2));
  OutputParticle(-PDGSCL,model->displayPhys().mu.display(1,2));
  OutputParticle(-PDGSB1,min(model->displayPhys().md.display(1,3),
                             model->displayPhys().md.display(2,3)));
  OutputParticle(-PDGST1,min(model->displayPhys().mu.display(1,3),
                             model->displayPhys().mu.display(2,3)));
  OutputParticle(PDGSDR,model->displayPhys().md.display(2,1));
  OutputParticle(PDGSUR,model->displayPhys().mu.display(2,1));
  OutputParticle(PDGSSR,model->displayPhys().md.display(2,2));
  OutputParticle(PDGSCR,model->displayPhys().mu.display(2,2));
  OutputParticle(PDGSB2,max(model->displayPhys().md.display(1,3),
                             model->displayPhys().md.display(2,3)));
  OutputParticle(PDGST2,max(model->displayPhys().mu.display(1,3),
                             model->displayPhys().mu.display(2,3)));
  OutputParticle(-PDGSDR,model->displayPhys().md.display(2,1));
  OutputParticle(-PDGSUR,model->displayPhys().mu.display(2,1));
  OutputParticle(-PDGSSR,model->displayPhys().md.display(2,2));
  OutputParticle(-PDGSCR,model->displayPhys().mu.display(2,2));
  OutputParticle(-PDGSB2,max(model->displayPhys().md.display(1,3),
                             model->displayPhys().md.display(2,3)));
  OutputParticle(-PDGST2,max(model->displayPhys().mu.display(1,3),
                             model->displayPhys().mu.display(2,3)));
  OutputParticle(PDGSEL,decays->displaySlMass()(1));
  OutputParticle(PDGSNUEL,decays->displaySnuMass()(1));
  OutputParticle(PDGSMUL,decays->displaySlMass()(2));
  OutputParticle(PDGSNUMUL,decays->displaySnuMass()(2));
  OutputParticle(PDGSTAU1,decays->displaySlMass()(3));
  OutputParticle(PDGSNUTAUL,decays->displaySnuMass()(3));
  OutputParticle(-PDGSEL,decays->displaySlMass()(1));
  OutputParticle(-PDGSNUEL,decays->displaySnuMass()(1));
  OutputParticle(-PDGSMUL,decays->displaySlMass()(2));
  OutputParticle(-PDGSNUMUL,decays->displaySnuMass()(2));
  OutputParticle(-PDGSTAU1,decays->displaySlMass()(3));
  OutputParticle(-PDGSNUTAUL,decays->displaySnuMass()(3));

  OutputParticle(PDGSER,decays->displaySlMass()(4));
  OutputParticle(PDGSNUER,0);
  OutputParticle(PDGSMUR,decays->displaySlMass()(5));
  OutputParticle(PDGSNUMUR,0);
  OutputParticle(PDGSTAU2,decays->displaySlMass()(6));
  OutputParticle(PDGSNUTAUR,0);
  OutputParticle(-PDGSER,decays->displaySlMass()(4));
  OutputParticle(-PDGSNUER,0);
  OutputParticle(-PDGSMUR,decays->displaySlMass()(5));
  OutputParticle(-PDGSNUMUR,0);
  OutputParticle(-PDGSTAU2,decays->displaySlMass()(6));
  OutputParticle(-PDGSNUTAUR,0);
  OutputParticle(PDGGLUINO,model->displayPhys().mGluino);
  OutputParticle(PDGNEUTR1,model->displayPhys().mneut.display(1));
  OutputParticle(PDGNEUTR2,model->displayPhys().mneut.display(2));
  OutputParticle(PDGNEUTR3,model->displayPhys().mneut.display(3));
  OutputParticle(PDGNEUTR4,model->displayPhys().mneut.display(4));
  OutputParticle(PDGCHARGINO1,model->displayPhys().mch.display(1));
  OutputParticle(PDGCHARGINO2,model->displayPhys().mch.display(2));
  OutputParticle(-PDGCHARGINO1,model->displayPhys().mch.display(1));
  OutputParticle(-PDGCHARGINO2,model->displayPhys().mch.display(2));
  OutputParticle(PDGGRAVITINO,model->displayGravitino());
  OutputParticle(PDGTOP,model->displayDataSet().displayPoleMt());
  OutputParticle(-PDGTOP,model->displayDataSet().displayPoleMt());
  cout << "Output Particle Data: End" << endl;
}

//Calls all of the evaluation routines to evaluate the decay rates.  Using 
//the DecayCalc class output functions.
void DecayFormat::OutputDecayTable() {
  cout << "Decay Table: Start" << endl;
  decays->OutputNeutralinoDecays();
  cout << " Neutralino Decays: Done" << endl;
  decays->OutputCharginoDecays();
  cout << " Chargino Decays: Done" << endl;
  decays->OutputSleptonDecays();
  cout << " Slepton Decays: Done" << endl;
  decays->OutputSneutrinoDecays();
  cout << " Sneutrino Decays: Done" << endl;
  decays->OutputHiggsDecays();
  cout << " Higgs Decays: Done" << endl;
  cout << "Decay Table: End" << endl;
}

//Outputs the susy parameters in the order herwig expects.  This block of 
//parameters is transferred directly over to the ultimate Herwig input file.
//Note that the tanThetaTau parameter is now meaningless in the presence of
//general slepton mixing.
void DecayFormat::OutputSusyParams() {
  cout << "OutputSusyParams: Start" << endl;
  int i,j;
  double thetaR,thetaL;
  DoubleMatrix neutralinoMass(4,4);

  fout << setprecision(12);

  fout << model->displayTanb() << " ";
  fout << model->displayPhys().thetaH << " ";

  neutralinoMass = model->displayPhys().mixNeut 
    * DoubleMatrix(model->displayPhys().mneut) 
    * model->displayPhys().mixNeut.transpose();

  for(i=1;i<=4;i++) {
    for(j=1;j<=4;j++) {
      fout << model->displayPhys().mixNeut.display(j,i) << " ";
    }
    fout << endl;
  }
  thetaR = model->displayPhys().thetaR;
  thetaL = model->displayPhys().thetaL;
  fout << cos(thetaR) << " " << sin(thetaR) << " " << -sin(thetaR) << " "
       << cos(thetaR) << endl;
  fout << cos(thetaL) << " " << sin(thetaL) << " " << -sin(thetaL) << " "
       << cos(thetaL) << endl;

  if(model->displayPhys().mu.display(1,3) 
     < model->displayPhys().mu.display(2,3))
    fout << model->displayPhys().thetat << " ";
  else
    fout << PI/2.0 + model->displayPhys().thetat << " ";
  if(model->displayPhys().md.display(1,3) 
     < model->displayPhys().md.display(2,3))
    fout << model->displayPhys().thetab << " ";
  else
    fout << PI/2.0 + model->displayPhys().thetab << " ";
  if(model->displayPhys().me.display(1,3) 
     < model->displayPhys().me.display(2,3))
    fout << model->displayPhys().thetatau << endl;
  else
    fout << PI/2.0 + model->displayPhys().thetatau << endl;
  fout << model->displaySoftA(UA,3,3) << " ";
  fout << model->displaySoftA(DA,3,3) << " ";
  fout << model->displaySoftA(EA,3,3) << endl;
  fout << model->displaySusyMu() << endl;
  fout << "T" << endl;
  fout << model->displayDataSet().displayPoleMb() << endl;
  fout << model->displayMw() << endl;
  cout << "OutputSusyParams: End" << endl;
}

//Top-level control function to output the intermediate slepton decays file.
//This is the function called from the main program.
void DecayFormat::OutputIntermediateFile() {

  cout << endl << "Producing flavor violating slepton decay file"
		   << endl << endl;

  OutputParticleData();
  fout << "DECAYS:" << endl;
  OutputDecayTable();
  fout << "PARAMS:" << endl;
  OutputSusyParams();
  cout << "Finished producing flavor violating slepton decay file"
       << endl << endl;
  cout << "Now entering SUSYHIT program" << endl;
}
