///////////////////////////////////////////////////////////////////////////////
//                          Particle Data Classes
//
//    The classes contained here use combinations of previously generated files
//  to generate final Herwig, susy LesHouches, and SLHA2 format output files
//  from SPICE.
//
//                          DecayMode Class
//
//    The DecayMode class contains the information for a single decay.  The 
//  DecayMode class contains the parent and daughter particle Herwig Id's, 
//  the daughter particle PDG Id's (parent PDG Id's are unneeded because of 
//  the structure of SLHA files), the decay width, and the branching ratio.
//  The important methods are:
//
//    addDaughter              Handles the addition of a daughter particle
//                             to the decay mode
//
//    OutputHerwigDecayMode    Handles the output of this individual decay 
//    OutputSLHADecayMode      in a Herwig/SLHA format file.  These methods
//                             take as input the appropriate output stream
//                             and output a single line decay to the stream.
//
//    IsSleptonic              Method to check whether a decay contains any
//                             sleptons or sneutrinos as parent or daughter
//                             particles; used to discriminate the decays 
//                             taken from the flavor-violating and those taken
//                             from the susyhit output file.
//
//                          Particle Class
//
//    The Particle class contains all the decay modes and output methods for a 
//  single parent particle.  The particle class contains a vector of type 
//  DecayMode used to store the decays of the particle.  The important methods
//  are:
//
//    AddDecayMode             Adds a single decay mode to the vector of decay
//                             modes.  This method has error checking to 
//                             ensure the decays associated with the particle
//                             all have that particle as the parent.
//
//    OutputHerwigDecays       Handles the output of the entire block of decays
//    OutputSLHADecays         for this particle.  These methods take as input
//                             the appropriate output stream and output a 
//                             block of individual decay modes preceded by the
//                             appropriate header line.
//
//    FixWidth                 Method to take the current decays and produce
//                             branching ratios.  This method is only called 
//                             after all decay have been entered for a 
//                             particular particle.
//
//                          Model Class
//
//    The Model class handles the overall control of the decay table and 
//  interfaces with inputs directly from main.  This class contains a map 
//  of particles and a number of storage variable for model parameters used
//  for a Herwig input file.  The important methods are:
//
//    ReadIntermediateFile     Reads in and stores the information contained in
//                             the flavor-violating file produced by the 
//                             softsusy portion of the code.
//
//    ReadLesHouchesFile       Reads in and stores the information contained in
//                             the flavor-conserving file produced by susyhit
//
//    Output[...]File          Output to the given type of file.  Output to a 
//                             LesHouches file uses the specified input file 
//                             for all blocks besides the mass hierarchy and 
//                             the decay table.
//
//    Combine                  Method used to read in the intermediate file and
//                             the LesHouches file and merge the decay tables.
//
//    OnShell                  Method used to check whether a particular decay
//                             is on-shell before inputting it into the decay
//                             table.
//
///////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <iomanip>
#include <ParticleData.h>
#include <ParticleCodesFC.h>

using namespace std;

DecayMode::DecayMode() {}

//Add a daughter particle to the decay mode
void DecayMode::addDaughter(int daughterId, int daughterPDGId) {
  daughterIds.push_back(daughterId);
  daughterPDGIds.push_back(daughterPDGId);
}

void DecayMode::setDaughterId(int i, int newId) {
  if(i <= daughterIds.size())
    daughterIds[i] = newId;
}

void DecayMode::setDaughterPDGId(int i, int newId) {
  if(i <= daughterPDGIds.size())
    daughterPDGIds[i] = newId;
}

// Returns 1 if the decay mode involves sleptons and zero if it does not
// Also removes gravitino modes generated by susyhit
bool DecayMode::IsSleptonic() {

  int i, particleId;

  if((parentId >= HERSSELM) && (parentId <= HERSSNUTRBR)  
     || parentId == HERGRAVTINO)
    return 1;
  for(i = 0; i < daughterIds.size(); i++) {
    particleId = daughterIds[i];
    if((particleId >= HERSSELM) && (particleId <= HERSSNUTRBR ) 
       || particleId == HERGRAVTINO)
      return 1;
  }
  return 0;
}

// Outputs the decay mode in the format Herwig expects.
void DecayMode::OutputHerwigDecayMode(ostream& out) {
  int i;

  out << setiosflags(ios::scientific) << setprecision(8);
  out << setw(6) << parentId << setw(16) << branchingRatio << setw(6) << 0;
  for(i = 0; i < numDaughters; i++) {
    out << setw(6) << daughterIds[i];
  }
  for(i = numDaughters; i < 5; i++) {
    out << setw(6) << 0;
  }
  out << endl << resetiosflags(ios::scientific);
  return;
}

// Outputs the decay mode in the SLHA format.
void DecayMode::OutputSLHADecayMode(ostream& out) {
  int i;

  out << setiosflags(ios::scientific) << setprecision(8);
  out << setw(19) << branchingRatio;
  out << setw(5) << numDaughters << "  ";
  for(i = 0; i < numDaughters; i++){
    out << setw(10) << daughterPDGIds[i];
  }
  out << endl << resetiosflags(ios::scientific);
  return;
}

Particle::Particle() {
  herwigId = 0;
  mass = 0;
  totalDecayWidth = 0;
  numDecays = 0;
}

Particle::~Particle() {
  int i;
  for(i=0;i< decays.size();i++) {
    delete decays[i];
  }
}

//Add a decay mode to the vector of decay modes contained in the particle
void Particle::AddDecayMode(DecayMode* mode) {
  //Check to ensure all decays contained in a particle actually have that
  //particle as a parent
  if(herwigId == mode->getParentId()) {
     numDecays++;
     decays.push_back(mode);
  }
  else{
    cout<<"Error: attempting to insert the decay mode ";
    mode->OutputHerwigDecayMode(cout);
    cout<<"into the list of decay modes for parent particle "<<herwigId
	<<".  Discarding mode."<<endl;
  }
}

// Sums up the decay widths to find the total decay rate. Then goes over the
// individual modes and calculates their branching ratios. Decays whose rates
// are below the tolerance are not included and their rates are set to zero.
void Particle::FixWidth() {
  int i, j;
  bool refix = false;
  double dWidth = 0;

  for(i = 0;i<decays.size();i++) {
    dWidth += decays[i]->getDecayWidth();
  }
  setTotalDecayWidth(dWidth);
  for(i = 0;i<decays.size();i++) {
    decays[i]->setBranchingRatio(decays[i]->getDecayWidth()/dWidth);
    if(decays[i]->getBranchingRatio() < BRCUT && 
       decays[i]->getBranchingRatio() > 0) {
      cout << " Decay of " << decays[i]->getParentId() << " into";
      for(j = 0; j < decays[i]->getNumDaughters(); j++) {
	cout << " " << decays[i]->getDaughterId(j);
      }
      cout << " has branching fraction of " << decays[i]->getBranchingRatio()
	   << " - Removing" << endl;

      decays[i]->setDecayWidth(0);
      decays[i]->setBranchingRatio(0);
      numDecays--;
      refix = true;
    }
  }

  if(refix) FixWidth();
}

double Particle::getLifetimeSeconds() {
  if(totalDecayWidth != 0)
    return HBAR/totalDecayWidth;
  else
    return 0.1E31;
}

// Runs over all decay modes of a particle and outputs them.
void Particle::OutputHerwigDecays(ostream& out) {
  int i;

  out<<setw(4)<<numDecays<<endl;

  for(i=0;i<decays.size();i++) {
    if(decays[i]->getDecayWidth()!=0)
      decays[i]->OutputHerwigDecayMode(out);
  }
}

// Runs over all decay modes of a particle and outputs them in SLHA format.
void Particle::OutputSLHADecays(ostream& out) {
  int i;

  out << "DECAY" << setw(10) << PDGId
      << setiosflags(ios::scientific) << setprecision(8) 
      << setw(19) << getTotalDecayWidth() << endl; 
  out << "#          BR         NDA      ID1       ID2       ID3" << endl;
  out << resetiosflags(ios::scientific);
  for(i=0;i<decays.size();i++) {
    if(decays[i]->getDecayWidth()!=0 && PDGId > 0){
      decays[i]->OutputSLHADecayMode(out);
    }
  }
  return;
}


Model::Model():offshellmsg(true) {}

Model::~Model() {
  Particle* pParticle;
  map<int,Particle*>::iterator it;

  for(it=particleMap.begin();it!=particleMap.end();it++) {
    delete (*it).second;
  }
}

// Converts a PDG code particle into a Herwig Code
int Model::PDGToHerwig(int pdgId) {
  switch (abs(pdgId)) {
  case PDGDOWN: return ((pdgId > 0) ? HERDQRK : HERDBAR);
  case PDGUP: return ((pdgId > 0) ? HERUQRK : HERUBAR);
  case PDGSTRANGE: return ((pdgId > 0) ? HERSQRK : HERSBAR);
  case PDGCHARM: return ((pdgId > 0) ? HERCQRK : HERCBAR);
  case PDGBOTTOM: return ((pdgId > 0) ? HERBQRK : HERBBAR);
  case PDGTOP: return ((pdgId > 0) ? HERTQRK : HERTBAR);
  case PDGELECTRON: return ((pdgId > 0) ? HEREM : HEREP);
  case PDGENEUTRINO: return ((pdgId > 0) ? HERNU_E : HERNU_EBAR);
  case PDGMUON: return ((pdgId > 0) ? HERMUM : HERMUP);
  case PDGMUNEUTRINO: return ((pdgId > 0) ? HERNU_MU : HERNU_MUBAR);
  case PDGTAU: return ((pdgId > 0) ? HERTAUM : HERTAUP);
  case PDGTAUNEUTRINO: return ((pdgId > 0) ? HERNU_TAU : HERNU_TAUBR);
  case PDGGLUON: return HERGLUON;
  case PDGPHOTON: return HERGAMMA;
  case PDGZ: return HERZ0;
  case PDGW: return ((pdgId > 0) ? HERWP : HERWM);
  case PDGHIGGS: return HERHIGGSL0;
  case PDGHEAVYHIGGS: return HERHIGGSH0;
  case PDGCPODDHIGGS: return HERHIGGSA0;
  case PDGCHARGEDHIGGS: return ((pdgId > 0) ? HERHIGGSP : HERHIGGSM);
  case PDGGRAVITON: return HERGRAVITON;
  case PDGSDL: return ((pdgId > 0) ? HERSSDL : HERSSDLBR);
  case PDGSUL: return ((pdgId > 0) ? HERSSUL : HERSSULBR);
  case PDGSSL: return ((pdgId > 0) ? HERSSSL : HERSSSLBR);
  case PDGSCL: return ((pdgId > 0) ? HERSSCL : HERSSCLBR);
  case PDGSB1: return ((pdgId > 0) ? HERSSB1 : HERSSB1BR);
  case PDGST1: return ((pdgId > 0) ? HERSST1 : HERSST1BR);
  case PDGSDR: return ((pdgId > 0) ? HERSSDR : HERSSDRBR);
  case PDGSUR: return ((pdgId > 0) ? HERSSUR : HERSSURBR);
  case PDGSSR: return ((pdgId > 0) ? HERSSSR : HERSSSRBR);
  case PDGSCR: return ((pdgId > 0) ? HERSSCR : HERSSCRBR);
  case PDGSB2: return ((pdgId > 0) ? HERSSB2 : HERSSB2BR);
  case PDGST2: return ((pdgId > 0) ? HERSST2 : HERSST2BR);
  case PDGSEL: return ((pdgId > 0) ? HERSSELM : HERSSELP);
  case PDGSNUEL: return ((pdgId > 0) ? HERSSNUEL : HERSSNUELBR);
  case PDGSMUL: return ((pdgId > 0) ? HERSSMULM : HERSSMULP);
  case PDGSNUMUL: return ((pdgId > 0) ? HERSSNUMUL : HERSSNUMLBR);
  case PDGSTAU1: return ((pdgId > 0) ? HERSSTAU1M : HERSSTAU1P);
  case PDGSNUTAUL: return ((pdgId > 0) ? HERSSNUTL : HERSSNUTLBR);
  case PDGSER: return ((pdgId > 0) ? HERSSERM : HERSSERP);
  case PDGSNUER: return ((pdgId > 0) ? HERSSNUER : HERSSNUERBR);
  case PDGSMUR: return ((pdgId > 0) ? HERSSMURM : HERSSMURP);
  case PDGSNUMUR: return ((pdgId > 0) ? HERSSNUMUR : HERSSNUMRBR);
  case PDGSTAU2: return ((pdgId > 0) ? HERSSTAU2M : HERSSTAU2P);
  case PDGSNUTAUR: return ((pdgId > 0) ? HERSSNUTR : HERSSNUTRBR);
  case PDGGLUINO: return HERGLUINO;
  case PDGNEUTR1: return HERNTLINO1;
  case PDGNEUTR2: return HERNTLINO2;
  case PDGNEUTR3: return HERNTLINO3;
  case PDGNEUTR4: return HERNTLINO4;
  case PDGCHARGINO1: return ((pdgId > 0) ? HERCHGINO1P : HERCHGINO1M);
  case PDGCHARGINO2: return ((pdgId > 0) ? HERCHGINO2P : HERCHGINO2M);
  case PDGGRAVITINO: return HERGRAVTINO;
  }
}

Particle* Model::FindParticle(int id) {
  return particleMap.find(id)->second;
}

//Check whether a particular decay mode is on shell
bool Model::OnShell(DecayMode *decay) {
  int i, daughterId, parentId, numDaughters;
  double totalMass, parentMass;
  
  totalMass = 0;
  
  parentId = decay->getParentId();
  parentMass = fabs(FindParticle(parentId)->getMass());

  numDaughters = decay->getNumDaughters();

  for(i = 0; i < numDaughters; i++) {
    daughterId = decay->getDaughterId(i);
    switch(daughterId) {
    case HERDQRK: totalMass += MD;  break;
    case HERUQRK: totalMass += MU;  break;
    case HERSQRK: totalMass += MS;  break;
    case HERCQRK: totalMass += MC;  break;
    case HERBQRK: totalMass += MB;  break;
    case HERTQRK: totalMass += MT;  break;
    case HERDBAR: totalMass += MD;  break;
    case HERUBAR: totalMass += MU;  break;
    case HERSBAR: totalMass += MS;  break;
    case HERCBAR: totalMass += MC;  break;
    case HERBBAR: totalMass += MB;  break;
    case HERTBAR: totalMass += MT;  break;
    case HERGLUON: totalMass += 0;  break;
    case HERGAMMA: totalMass += 0;  break;
    case HEREM: totalMass += ME;  break;
    case HERNU_E: totalMass += 0;  break;
    case HERMUM: totalMass += MMU;  break;
    case HERNU_MU: totalMass += 0;  break;
    case HERTAUM: totalMass += MTAU;  break;
    case HERNU_TAU: totalMass += 0;  break;
    case HEREP: totalMass += ME;  break;
    case HERNU_EBAR: totalMass += 0;  break;
    case HERMUP: totalMass += MMU;  break;
    case HERNU_MUBAR: totalMass += 0;  break;
    case HERTAUP: totalMass += MTAU;  break;
    case HERNU_TAUBR: totalMass += 0;  break;
    case HERWP: totalMass += MW;  break;
    case HERWM: totalMass += MW;  break;
    case HERZ0: totalMass += MZ;  break;
    default: totalMass += fabs(FindParticle(daughterId)->getMass());
      break;
    }
  }
  if (totalMass > parentMass) {
    if(offshellmsg){
      cout << "Note: SUSYHIT calculates off shell Higgs "
           << "decays, which are truncated by SPICE." << endl;
      offshellmsg = false; // Turn off message
    }
    cout << " Off shell Decay of " << parentId << " Into";
    for(i=0;i<numDaughters;i++) {
      cout << " " << decay->getDaughterId(i);
    }
    cout << " - Removing" << endl;
    return 0;
  }
  else
    return 1;
}

// Reads in the Herwiglike file generated by softsusy. For each particle
// creates a particle object, and for each decay inserts it into the particle 
// decay list.  Also, for each decay entered, the charged conjugate decay 
//is also created.
void Model::ReadIntermediate(ifstream& fin) {
  cout << "Reading Intermediate File: Start" << endl;

  int numParticles, i, j, parentId, parentPDGId, numDaughters, daughterId, 
    daughterPDGId;
  double mass, width, totalWidth;
  char temp[100];
  DecayMode* decay;
  Particle *particle;

  //Read in the mass hierarchy
  fin >> numParticles;
  for (i = 1 ; i <= numParticles; i++) {
    particle = new Particle;
    fin >> parentPDGId >> mass;
  
    //Generate a particle with the given Id, record its properties with an
    //initial decay width set to zero.
    parentId = PDGToHerwig(parentPDGId);
    particle->setHerwigId(parentId);
    particle->setPDGId(parentPDGId);
    particle->setMass(mass);
    particle->setTotalDecayWidth(0);
    particleMap[parentId]=particle;
  }

  fin >> temp;
  fin >> temp;

  //Read in the decay table
  while (strcmp(temp,"PARAMS:")) {
    decay = new DecayMode;
    fin >> parentId >> width >> numDaughters;
    decay->setDecayWidth(width);

    if(totalWidth != 0)
      decay->setBranchingRatio(width/totalWidth);

    decay->setNumDaughters(numDaughters);
    parentId = PDGToHerwig(parentId);
    decay->setParentId(parentId);

    for(i = 1;i<=numDaughters;i++) {
      fin >> daughterPDGId;
      daughterId = PDGToHerwig(daughterPDGId);
      decay->addDaughter(daughterId, daughterPDGId);
    }

    if(width!=0)
      particleMap[parentId]->AddDecayMode(decay);

    fin >> temp;
  }

  //Read in the Herwig file model parameters
  fin >> tanBeta >> alphaH;
  for(i=1;i<= 16;i++) {
    fin >> neutralinoMix[i-1];
  }
  for(i=1;i<=4;i++)
    fin >> charginoV[i-1];
  for(i=1;i<=4;i++)
    fin >> charginoU[i-1];
  fin >> thetaTop >> thetaBottom >> thetaTau >> aTop >> aBottom >> aTau >> mu;
  fin >> rparity;
  fin >> bottomMass >> WpMass;
  
  cout << "Reading Intermediate File: End" << endl;
}

// Reads in the les houches file generated by susy hit. Ignores sleptonic
// decays since these are generated by the softsusy program. Also, add in the
// charged conjugate mode.
void Model::ReadLesHouchesFile(ifstream& fin) {
  cout << "Reading Les Houches File: Start" << endl;

  char strTemp[300];
  int i, numDaughters, parentPDGId, parentId, parentCCId, daughterPDGId, 
    daughterId, daughterCCId, done;
  double width, br;
  Particle *particle,*particleCC;
  DecayMode *decay, *decayCC;
  map<int,Particle*>::iterator it, itCC;

  while(!fin.eof()) {

    //Discard everything in the file up to the decay table
    if(!strcmp(strTemp,"DECAY")) {

      //Read in the header line of the decay block and either locate the 
      //particle associated with or create a new particle.
      fin >> parentPDGId >> width;

      parentId = PDGToHerwig(parentPDGId);
      parentCCId = PDGToHerwig(-parentPDGId);
      it = particleMap.find(parentId);
      itCC = particleMap.find(parentCCId);

      if(it == particleMap.end()) {
        particle = new Particle;
        particle->setHerwigId(parentId);
        particleMap[parentId] = particle;
      }
      else
        particle =  particleMap[parentId];

      if(parentCCId != parentId) {
        if(itCC == particleMap.end()) {
          particleCC = new Particle;
          particleCC->setHerwigId(parentCCId);
          particleMap[parentCCId] = particleCC;
        }
        else
          particleCC=particleMap[parentCCId];
      }

      //If there are non-zero decay modes associated with the particle, 
      //begin reading them in
      if(width != 0) {
        fin.getline(strTemp,300);
        fin.getline(strTemp,300);
        done = fin.peek();
        if((char)done == '#') {
          fin.getline(strTemp,300);
          cout<<strTemp<<endl;
          done = ' ';
        }

        while((char)done != '#' && !fin.eof()) {
          decay = new DecayMode;
          decayCC = new DecayMode;
          fin >> br >> numDaughters;
          decay->setNumDaughters(numDaughters);
          decay->setDecayWidth(br*width);
          decay->setParentId(parentId);
          decayCC->setNumDaughters(numDaughters);
          decayCC->setDecayWidth(br*width);
          decayCC->setParentId(parentCCId);
          for(i = 1; i <= numDaughters; i++) {
            fin >> daughterPDGId;
            daughterId = PDGToHerwig(daughterPDGId);
            decay->addDaughter(daughterId, daughterPDGId);
            daughterCCId = PDGToHerwig(-daughterPDGId);
            decayCC->addDaughter(daughterCCId, -daughterPDGId);
          }
          if(!decay->IsSleptonic() && OnShell(decay)) {
            particle->AddDecayMode(decay);
            if(parentCCId != parentId)
              particleCC->AddDecayMode(decayCC);
          }
          else {
            delete decay;
          }
          fin.getline(strTemp,300);
          done = fin.peek();
          if((char)done == '#' || done == char_traits<char>::eof()) {
            if(done == char_traits<char>::eof())
              fin >> done;
            particle->FixWidth();
            particleCC->FixWidth();
          }
        }
      }
      else {
        particle->FixWidth();
        particleCC->FixWidth();
      }
    }
    fin >> strTemp;
  }

  cout << "Reading Les Houches File: End" << endl;
}

void Model::Combine(ifstream &intin, ifstream &slhain) {
  ReadIntermediate(intin);
  ReadLesHouchesFile(slhain);
}

// Outputs the file in the format Herwig expects.
void Model::OutputHerwigFile(ostream& out) {

  cout << "Output Herwig File: Start" << endl;

  out<<setw(4)<<65<<endl;

  map<int, Particle*>::iterator it;
  Particle *particle;
  int i;

  for(it=particleMap.begin();it!=particleMap.end();it++) {
    particle = it->second;

    out<<setw(5)<<particle->getHerwigId()<<setiosflags(ios::fixed)
        <<setw(12)<<setprecision(4)<<particle->getMass()
        <<resetiosflags(ios::fixed)<<setiosflags(ios::scientific)
        <<setw(15)<<setprecision(5)<<particle->getLifetimeSeconds()
        <<resetiosflags(ios::scientific)<<endl;
  }

  for(it=particleMap.begin();it!=particleMap.end();it++) {
    particle = it->second;
    if(particle->getHerwigId() == 6 || particle->getHerwigId() == 12) {
      out<<setw(4)<<0<<endl;;
    }
    else {
      particle->OutputHerwigDecays(out);
    }
  }

  out<<setiosflags(ios::fixed)<<setprecision(8);
  out<<setw(16)<<tanBeta<<setw(16)<<alphaH<<endl;
  for(int i=0;i<4;i++) {
    for(int j=0;j<4;j++) {
      out<<setw(16)<<neutralinoMix[4*i+j];
    }
    out<<endl;
  }
  for(int i=0;i<4;i++)
    out<<setw(16)<<charginoV[i];
  out<<endl;
  for(int i=0;i<4;i++)
    out<<setw(16)<<charginoU[i];
  out<<endl;
  out<<setw(16)<<thetaTop<<setw(16)<<thetaBottom<<setw(16)<<thetaTau<<endl;
  out<<setw(16)<<aTop<<setw(16)<<aBottom<<setw(16)<<aTau<<endl;
  out<<setw(16)<<mu<<endl<<setw(5)<<rparity<<endl;

  cout << "Output Herwig File: End" << endl;
}

// Outputs the file in Les Houches Accord format
void Model::OutputLesHouchesFile(ifstream &fin, ostream &out, int type) {
  cout << "Output Les Houches File: Start" << endl;

  LesHouchesHeader(out, type);

  map<int, Particle*>::iterator it;
  Particle *particle;
  int i;

  string temp;
  while(getline(fin,temp)) {
    // break loop on beginning of decay table
    if(!strcmp(temp.c_str(),"#         PDG            Width")) break;
    // replace mass block with corrected one
    if(!strcmp(temp.c_str(),"BLOCK MASS  # Mass Spectrum")) {
      out << temp << endl;
      out << "# PDG code           mass       particle" << endl;
      out << setiosflags(ios::scientific) << setprecision(8);
      out << "        24     " << WpMass << endl;
      out << "         5     " << bottomMass 
          << "   # b-quark pole mass calculated from mb(mb)_Msbar" << endl;

      for(it=particleMap.begin();it!=particleMap.end();it++) {
        particle = it->second;
        if(particle->getPDGId() < 0) continue;
        out << setw(10) << particle->getPDGId() 
            << setw(19) << particle->getMass() << endl;
      } 
      out << resetiosflags(ios::scientific);     
      while(getline(fin,temp))
        if(!strcmp(temp.c_str(),"#")) break;
      out << temp << endl;
      continue;
    }
    if(temp[0] != '#')
      out << temp << endl;
  }

  for(it=particleMap.begin();it!=particleMap.end();it++) {
    particle = it->second;
    if(particle->getPDGId() < 0) continue;
    if(particle->getTotalDecayWidth() == 0) continue;
    out << "#         PDG            Width" << endl;
    particle->OutputSLHADecays(out);
  }
  cout << "Output Les Houches File: End" << endl;
  return;
}

void Model::LesHouchesHeader(ostream &out, int type) {

  //Spice header
  out<<"#\n"
     <<"#                          ======================\n"
     <<"#                          |    SPICE OUTPUT    |\n"
     <<"#                          ======================\n"
     <<"#\n"
     <<"#          ------------------------------------------------------\n";
  if(type == 1) 
    out<<"#          |         Output in LesHouches Accord Format         |\n";
  if(type == 2)
    out<<"#          |        Output in LesHouches Accord 2 Format        |\n";
  out<<"#          ------------------------------------------------------\n"
     <<"#\n"
     <<"#          ------------------------------------------------------\n"
     <<"#          |    Created by G. Engelhard, J. Feng, I. Galon,     |\n"
     <<"#          |  D. Sanford, and F. Yu.  For problems with SPICE   |\n"
     <<"#          |  send e-mail to                                    |\n"
     <<"#          |          spice@hep.ps.uci.edu                      |\n"
     <<"#          |                                                    |\n"
     <<"#          |    If you use SPICE, please cite                   |\n"
     <<"#          |      arXiv:0904.1415v1 [hep-ph]                    |\n"
     <<"#          |                                                    |\n"
     <<"#          |    If your model involves flavor-violating slepton |\n"
     <<"#          |      3-body decays of sleptons, please cite        |\n"
     <<"#          |      arXiv:0904.1416v1 [hep-ph]                    |\n"
     <<"#          |                                                    |\n"
     <<"#          |    We reprint the original SOFTSUSY and SUSYHIT    |\n"
     <<"#          |  headers below.  Please note that we use SOFTSUSY  |\n"
     <<"#          |  for spectrum generation and SUSYHIT for non-      |\n"
     <<"#          |  leptonic decay calculations, and thus comments or |\n"
     <<"#          |  questions about these calculation processes       |\n"
     <<"#          |  should be directed to the appropriate authors.    |\n"
     <<"#          ------------------------------------------------------\n"
     <<"#\n"
     <<"#\n";

  //Copy of softsusy header
  out<<"#\n"
     <<"#                          ======================\n"
     <<"#                          |   SOFTSUSY HEADER  |\n"
     <<"#                          ======================\n"
     <<"#\n"
     <<"# SOFTSUSY3.0.2\n"
     <<"# B.C. Allanach, Comput. Phys. Commun. 143 (2002) 305-331, "
     <<"hep-ph/0104145\n"
     <<"#\n"
     <<"#\n";

  //Copy of the susyhit header
  out<<"#\n"
     <<"#                          ======================\n"
     <<"#                          |   SUSYHIT HEADER   |\n"
     <<"#                          ======================\n"
     <<"#\n"
     <<"#\n"
     <<"#          ------------------------------------------------------\n"
     <<"#          |     This is the output of the SUSY-HIT package     |\n"
     <<"#          |  created by A.Djouadi, M.Muehlleitner and M.Spira. |\n"
     <<"#          |  In case of problems with SUSY-HIT email to        |\n"
     <<"#          |           margarete.muehlleitner@cern.ch           |\n"
     <<"#          |           michael.spira@psi.ch                     |\n"
     <<"#          |           abdelhak.djouadi@cern.ch                 |\n"
     <<"#          ------------------------------------------------------\n"
     <<"#\n"
     <<"#          ------------------------------------------------------\n"
     <<"#          |  SUSY Les Houches Accord - MSSM Spectrum + Decays  |\n"
     <<"#          |              based on the decay programs           |\n"
     <<"#          |                                                    |\n"
     <<"#          |                     SDECAY 1.3                     |\n"
     <<"#          |                                                    |\n"
     <<"#          |  Authors: M.Muhlleitner, A.Djouadi and Y.Mambrini  |\n"
     <<"#          |  Ref.:    Comput.Phys.Commun.168(2005)46           |\n"
     <<"#          |           [hep-ph/0311167]                         |\n"
     <<"#          |                                                    |\n"
     <<"#          |                     HDECAY 3.302                   |\n"
     <<"#          |                                                    |\n"
     <<"#          |  By: A.Djouadi,J.Kalinowski,M.Muhlleitner,M.Spira  |\n"
     <<"#          |  Ref.:    Comput.Phys.Commun.108(1998)56           |\n"
     <<"#          |           [hep-ph/9704448]                         |\n"
     <<"#          |                                                    |\n"
     <<"#          |                                                    |\n"
     <<"#          |  If not stated otherwise all DRbar couplings and   |\n"
     <<"#          |  soft SUSY breaking masses are given at the scale  |\n"
     <<"#          |  Q=  0.97084150E+03                                |\n"
     <<"#          |                                                    |\n"
     <<"#          ------------------------------------------------------\n"
     <<"#\n"
     <<"#\n";
}
