#ifndef PARTICLE_DATA_H
#define PARTICLE_DATA_H
#include <vector>
#include <map>
#include <fstream>
#include <math.h>

#define HBAR 6.58211915E-25 // hbar
#define BRCUT 1e-20         //Cut on very small branching ratios

// Namespaces
using std::map;
using std::vector;
using std::fstream;
using namespace std;

// A class encoding a single decay mode. Contains the parent, a vector of the
// daughter particles, the decay rate and the branching ratio.
class DecayMode {
private:
  static double brCut;                 //Cut on very small branching ratios
  double decayWidth, branchingRatio;   //The decay width and branching ratio
  int numDaughters, parentId;          //Number of daughter particles and the 
                                       //Herwig Id of the parent particles
  vector<int> daughterIds;     //Herwig Id's of daughter particles
  vector<int> daughterPDGIds;  //PDG Id's of daughter particles

public:
  DecayMode();

  //Get the values of stored variables
  double getDecayWidth() {return decayWidth;}
  double getBranchingRatio() {return branchingRatio;}
  int getParentId() {return parentId;}
  int getNumDaughters() {return numDaughters;}
  int getDaughterId(int i) {return daughterIds[i];}
  int getDaughterPDGId(int i) {return daughterPDGIds[i];}

  bool IsSleptonic();  // returns true if the decay contains leptons either 
                       //as a parent or a daughter.

  //Alter values of stored variables
  void setDecayWidth(double width) {decayWidth = width;}
  void setNumDaughters(int n) {numDaughters = n;}
  void setParentId(int id) {parentId = id;}
  void setBranchingRatio(double br) {branchingRatio = br;}
  void setDaughterId(int i,int newId);
  void setDaughterPDGId(int i, int newId);

  //Add a daughter particle to the decay
  void addDaughter(int daughterId, int daughterPDGId);

  //Output the decay mode formatted for either a Herwig or SLHA file
  void OutputHerwigDecayMode(ostream& out);
  void OutputSLHADecayMode(ostream& out);
};

// Encodes all decay modes for a single particle. Contains the mass of the
// particle, and a list of its decay modes.
class Particle {
 private:
  int herwigId;
  int PDGId;
  double mass;
  double totalDecayWidth;
  int numDecays;
  vector<DecayMode*> decays;

 public:
  Particle();
  ~Particle();

  //Get the values of stored variables
  int getHerwigId() {return herwigId;}
  int getPDGId() {return PDGId;}
  double getMass() {return mass;}
  double getTotalDecayWidth() {return totalDecayWidth;}
  int getNumDecays() {return numDecays;}
  double getLifetimeSeconds();
  vector<DecayMode*>& getDecayVector() {return decays;}

  //Set the values of stored variables
  void setHerwigId(int id) {herwigId = id;}
  void setPDGId(int id) {PDGId = id;}
  void setMass(double m) {mass = m;}
  void setTotalDecayWidth(double width) {totalDecayWidth = width;}
  void setNumDecays(int n) {numDecays = n;}

  //Output a block of all decays for this particle formatted for either 
  //a Herwig or SLHA file 
  void OutputHerwigDecays(ostream& out);
  void OutputSLHADecays(ostream &out);

  //Add a new decay mode for this particles
  void AddDecayMode(DecayMode* mode);

  //Use the current list of decay widths to calculate the branching ratios
  void FixWidth();
};

// Encodes an MSSM model. Contains the model parameters to be outputted to
// Herwig, as well as a list of the susy particles (together with their decay
// modes).  Note that it is a singleton class. Finally, when each particle is
// read, a call to FixWidth is made to set the corrected branching ratios.
class Model {
private:

  //Storage map containing the particles
  map<int,Particle*> particleMap;

  //Variables for the final herwig file
  double tanBeta, alphaH;
  double neutralinoMix[16], charginoV[4], charginoU[4];
  double thetaTop, thetaBottom, thetaTau, aTop, aBottom, aTau, mu;
  char rparity;
  double bottomMass, WpMass;

  //Convert a PDG Id to a Herwig Id
  int PDGToHerwig(int pdgId);

  //Check whether a particular decay width is on shell (run on input)
  bool OnShell(DecayMode *decay);
  bool offshellmsg;

  //Read in the intermediate file of flavor-violating decays file produced 
  //by the softsusy portion of the code
  void ReadIntermediate(ifstream &in);

  //Read in the file of flavor-conserving decays produced by susyhit
  void ReadLesHouchesFile(ifstream &in);

public:
  Model();
  ~Model();

  //Function to combine the input flavor-violating and flavor-conserving 
  //decay files.  Creates a table of decay widths.
  void Combine(ifstream &intin, ifstream &slhain);

  //Output various types of files.  The LesHouches accord output files use the
  //input SLHA or SLHA2 file for output except the mass hierarchy and decays.
  void OutputHerwigFile(ostream &out);
  void OutputLesHouchesFile(ifstream &in, ostream &out, int type);
  void LesHouchesHeader(ostream &out, int type);

  //Get a particular particle out of the particle map from its id
  Particle* FindParticle(int id);
};

#endif
