///////////////////////////////////////////////////////////////////////////////
//                           ThreeBodyIntegrals Class
//    This class handles the phase space integrals found in slepton three-body
//  decays.  The integration is performed by using a quadratic Simpson's rule
//  implementation (method found at the end of the file for convenience).
//  The masses of parent, daughter, and intermediate particles are defined
//  using the method setMasses, which defines the dimensionless values
//  r_{particle} = m_{particle}/m_i, the starting and ending points of 
//  integration, and the common factor R2.
//    Generically, the three-body phase space can be easily reduced to a 
//  two-dimensional integral over the energies of the two final-state fermions.
//  For numerical simplicity we have performed the first integral analytically;
//  the resulting one-dimensional integral is best performed numerically.  
//  The methods FirstUpperLimit and FirstLowerLimit provide the (dimensionless)
//  limits of this first integral.  The full contribution of this first 
//  integral is given in the functions F, F1, F2, etc; the full form of these 
//  functions is complicated, but in general they represent integrations over 
//  the first fermion energy with various propagator structures and numerators
//  constant or linear in the fermion energy.
//    There are three general classes of methods in this class:
//
//  [...]Integral#          The individual expressions for integrals over 
//                          phase space for a given combination of 
//                          intermediate off-shell particles; the full 
//                          expressions forms are given in the appendix 
//                          of ref. [] in the manual.  Each integral is 
//                          dimensionless and a function only of the 
//                          dimensionless variable x = 2 E_k/m_i and 
//                          various particle masses.  In order to allow for 
//                          natural passing of these methods to the Simpson's
//                          integration method, these methods and all 
//                          variables and methods they depend on are defined
//                          as static - in particular the masses are 
//                          statically defined and thus multiple copies of the
//                          class will always possess possess the same values
//                          for the masses.
//
//  evaluate[...]Integrals  Methods which sum the contributions of a certain 
//                          set of integrals with given weights.  
//                          These methods are passed either a single 
//                          double value or a vector of double values (based 
//                          on the number of unique integrals associated with
//                          a particular pairing of intermediate particles),
//                          which then weight the integrals that are only 
//                          dependent on the given particle masses.  This 
//                          separates the phase space contributions (which are
//                          the same from mode to mode up to redefinition of 
//                          masses) and the integral weights (which vary 
//                          significantly from mode to mode, and are in any 
//                          case more easily calculated in DecayCalc.  It also
//                          ensures that the masses and integration limits are
//                          kept constant between different integrals of the
//                          same type.
//
//  A function call to ThreeBodyIntegrals requires:
//    a) Definition of a coefficient or vector of coefficients
//    b) A initial slepton mass and a vector of 5 other masses
//       (the final slepton mass, the two fermion, and the two intermediate
//       particles in that order) which will be passed to the setMasses method
//    c) A call to the appropriate evaluate[...]Integrals method with the
//       coefficient or vector of coefficients
//
///////////////////////////////////////////////////////////////////////////////

#include<math.h>
#include<iostream>
#include<utils.h>
#include<ThreeBodyIntegrals.h>

//Set all static variable to zero initially
double ThreeBodyIntegrals::start = 0;
double ThreeBodyIntegrals::end = 0;
double ThreeBodyIntegrals::rslj = 0;
double ThreeBodyIntegrals::rk = 0;
double ThreeBodyIntegrals::rm = 0;
double ThreeBodyIntegrals::ra = 0;
double ThreeBodyIntegrals::rb = 0;
double ThreeBodyIntegrals::R2 = 0;

//Default constructor
ThreeBodyIntegrals::ThreeBodyIntegrals() {}

//This function takes the parent particle mass m0, a vector of 5 other masses 
//(the final slepton mass, the two fermion, and the two intermediate particles 
//in that order).  From these the dimensionless variables the methods sets 
//r_{particle} = m_{particle}/m_i, as well as the integration limits.
void ThreeBodyIntegrals::setMasses(DoubleVector& masses, double m0) {
  rslj = masses(1) / m0;
  rk = masses(2) / m0;
  rm = masses(3) / m0;
  ra = masses(4) / m0;
  rb = masses(5) / m0;
  start = 2 * rk;
  end = 1 + sqr(rk) - sqr(rslj + rm);
  R2 = 1 - sqr(rslj) + sqr(rk) + sqr(rm);
}

/////////////////////////// Evaluation Functions //////////////////////////////

double ThreeBodyIntegrals::evaluateNeutralinoIntegrals(DoubleVector& c) {
  double sum = 0;

  sum += c(1) * Simpsons(ThreeBodyIntegrals::FermionIntegral1);
  sum += c(2) * Simpsons(ThreeBodyIntegrals::FermionIntegral2);
  sum += c(3) * Simpsons(ThreeBodyIntegrals::FermionIntegral3);
  sum += c(4) * Simpsons(ThreeBodyIntegrals::FermionIntegral4);
  sum += c(5) * Simpsons(ThreeBodyIntegrals::FermionIntegral5);
  sum += c(6) * Simpsons(ThreeBodyIntegrals::FermionIntegral6);

  return sum;
}

double ThreeBodyIntegrals::evaluateCharginoIntegral(double c) {
  return c * Simpsons(ThreeBodyIntegrals::FermionIntegral1);
}

double ThreeBodyIntegrals::evaluateScalarIntegral(double c) {
  return c * Simpsons(ThreeBodyIntegrals::ScalarIntegral);
}

double ThreeBodyIntegrals::evaluatePseudoscalarIntegral(double c) {
  return c * Simpsons(ThreeBodyIntegrals::PseudoscalarIntegral);
}

double ThreeBodyIntegrals::evaluateGaugeIntegrals(DoubleVector& c) {
  double sum = 0;

  sum += c(1) * Simpsons(ThreeBodyIntegrals::GaugeIntegral1);
  sum += c(2) * Simpsons(ThreeBodyIntegrals::GaugeIntegral2);

  return sum;
}

double ThreeBodyIntegrals::evaluateNeutralinoCrossIntegrals(DoubleVector& c) {
  double sum = 0;

  sum += c(1) * Simpsons(ThreeBodyIntegrals::FermionFermionCrossIntegral1);
  sum += c(2) * Simpsons(ThreeBodyIntegrals::FermionFermionCrossIntegral2);
  sum += c(3) * Simpsons(ThreeBodyIntegrals::FermionFermionCrossIntegral3);
  sum += c(4) * Simpsons(ThreeBodyIntegrals::FermionFermionCrossIntegral4);
  sum += c(5) * Simpsons(ThreeBodyIntegrals::FermionFermionCrossIntegral5);
  sum += c(6) * Simpsons(ThreeBodyIntegrals::FermionFermionCrossIntegral6);
  sum += c(7) * Simpsons(ThreeBodyIntegrals::FermionFermionCrossIntegral7);
  sum += c(8) * Simpsons(ThreeBodyIntegrals::FermionFermionCrossIntegral8);

  return sum;
}

double ThreeBodyIntegrals::evaluateScalarScalarCrossIntegral(double c) {
  return c * Simpsons(ThreeBodyIntegrals::ScalarScalarCrossIntegral);
}

double 
ThreeBodyIntegrals::evaluatePseudoscalarPseudoscalarCrossIntegral(double c) {
  return c*Simpsons(ThreeBodyIntegrals::PseudoscalarPseudoscalarCrossIntegral);
}

double ThreeBodyIntegrals::evaluateScalarFermionCrossIntegrals(DoubleVector& c)
{
  double sum = 0;

  sum += c(1) * Simpsons(ThreeBodyIntegrals::ScalarFermionCrossIntegral1);
  sum += c(2) * Simpsons(ThreeBodyIntegrals::ScalarFermionCrossIntegral2);

  return sum;
}

double 
ThreeBodyIntegrals::evaluatePseudoscalarFermionCrossIntegrals(DoubleVector& c){
  double sum = 0;

  sum +=c(1) * Simpsons(ThreeBodyIntegrals::PseudoscalarFermionCrossIntegral1);
  sum +=c(2) * Simpsons(ThreeBodyIntegrals::PseudoscalarFermionCrossIntegral2);

  return sum;
}

double ThreeBodyIntegrals::evaluateScalarGaugeCrossIntegral(double c) {
  return c * Simpsons(ThreeBodyIntegrals::ScalarGaugeCrossIntegral);
}

double ThreeBodyIntegrals::evaluateA0GaugeCrossIntegral(double c) {
  return c * Simpsons(ThreeBodyIntegrals::A0GaugeCrossIntegral);
}

double ThreeBodyIntegrals::evaluateG0GaugeCrossIntegral(double c) {
  return c * Simpsons(ThreeBodyIntegrals::G0GaugeCrossIntegral);
}

double ThreeBodyIntegrals::evaluateGaugeFermionCrossIntegrals(DoubleVector& c)
{
  double sum = 0;

  sum += c(1) * Simpsons(ThreeBodyIntegrals::GaugeFermionCrossIntegral1);
  sum += c(2) * Simpsons(ThreeBodyIntegrals::GaugeFermionCrossIntegral2);
  sum += c(3) * Simpsons(ThreeBodyIntegrals::GaugeFermionCrossIntegral3);
  sum += c(4) * Simpsons(ThreeBodyIntegrals::GaugeFermionCrossIntegral4);

  return sum;
}

double ThreeBodyIntegrals::evaluateGaugeCharginoCrossIntegral(double c) {
  return c * Simpsons(ThreeBodyIntegrals::GaugeFermionCrossIntegral1);
}

//////////////////////////// Phase Space Functions ////////////////////////////

double ThreeBodyIntegrals::Lambda(double x, double y, double z) {
  double lambda = (sqr(x) + sqr(y) + sqr(z) - 2*x*y - 2*x*z - 2*y*z);

  if(lambda < 0) {
    //Cut on small values of negative Lambda - used to protect against machine 
    //imprecision
    if(fabs(lambda) < 1.0e-12) return 1e-16;
    //Output a warning if the negative value is not sufficiently close to zero
    else cout << "Warning: lambda value less than 0 found: " << lambda << endl;
  }

  return lambda;
}

double ThreeBodyIntegrals::FirstUpperLimit(double x) {
  double lambda = Lambda(1 - x + sqr(rk), sqr(rm), sqr(rslj));

  return ((2 - x) * (R2 - x) + sqrt((sqr(x) - 4.0 * sqr(rk)) * lambda))
    / (2.0 * (1 - x + sqr(rk)));
}

double ThreeBodyIntegrals::FirstLowerLimit(double x) {
  double lambda = Lambda(1 - x + sqr(rk), sqr(rm), sqr(rslj));

  return ((2 - x) * (R2 - x) - sqrt((sqr(x) - 4.0 * sqr(rk)) * lambda))
    / (2.0 * (1 - x + sqr(rk)));
}

double ThreeBodyIntegrals::K(double x) {
  return (FirstUpperLimit(x) - FirstLowerLimit(x))
    / (1. - x + sqr(rk) - sqr(rb));
}

double ThreeBodyIntegrals::L(double x, double z) {
  return log(fabs( (FirstUpperLimit(x) + z)
		 / (FirstLowerLimit(x) + z) ));
}

double ThreeBodyIntegrals::F(double x) {
  return K(x) / ((1 - x + sqr(rk) - sqr(ra)) * (1 - x + sqr(rk)));
}

double ThreeBodyIntegrals::F1(double x) {
  return - L(x, sqr(ra) - 1 - sqr(rm))
    / (1 - x + sqr(rk) - sqr(rb));
}

double ThreeBodyIntegrals::F2(double x) {
  return F1(x) * (1 + sqr(rm) - sqr(ra)) - K(x);
}

double ThreeBodyIntegrals::G1(double x) {
  return L(x, x - 1 + sqr(rslj) - sqr(ra)) 
    / (1 - x + sqr(rk) - sqr(rb));
}

double ThreeBodyIntegrals::G2(double x) {
  return K(x) + (1 - x - sqr(rslj) + sqr(ra)) * G1(x);
}

double ThreeBodyIntegrals::H1(double x) {
  return (L(x, x - 1 + sqr(rslj) - sqr(ra))
	       - L(x, x - 1 + sqr(rslj) - sqr(rb))) / (sqr(ra) - sqr(rb));
}

double ThreeBodyIntegrals::H2(double x) {
  return L(x, x - 1 + sqr(rslj) - sqr(rb)) 
    + (1 - x - sqr(rslj) + sqr(ra)) * H1(x);
}

double ThreeBodyIntegrals::I1(double x) {
  return (FirstUpperLimit(x) - FirstLowerLimit(x)) 
    / ( (FirstUpperLimit(x) + x - 1 + sqr(rslj) - sqr(ra))
      * (FirstLowerLimit(x) + x - 1 + sqr(rslj) - sqr(ra)));
}

double ThreeBodyIntegrals::I2(double x) {
  return L(x, x - 1 + sqr(rslj) - sqr(ra)) 
    + (1 - x - sqr(rslj) + sqr(ra)) * I1(x);
}

///////////////////////////// Integration Methods /////////////////////////////

double ThreeBodyIntegrals::FermionIntegral1(double x) {
  return (x - 2 * sqr(rk)) * (1 - x + sqr(rk)) * (R2 - x) * F(x);
}

double ThreeBodyIntegrals::FermionIntegral2(double x) {
  return ra * rb * (x - 2 * sqr(rk)) * (R2 - x) * F(x);
}

double ThreeBodyIntegrals::FermionIntegral3(double x) {
  return 2.0 * rm * rb * (x - 2 * sqr(rk)) * (1 - x + sqr(rk)) * F(x);
}

double ThreeBodyIntegrals::FermionIntegral4(double x) {
  return 2.0 * rk * rb * (1 - x + sqr(rk)) * (R2 - x) * F(x);
}

double ThreeBodyIntegrals::FermionIntegral5(double x) {
  return 2.0 * rk * rm * ra * rb * (1 - x + sqr(rk)) * F(x);
}

double ThreeBodyIntegrals::FermionIntegral6(double x) {
  return 2.0 * rk * rm * sqr(1 - x + sqr(rk)) * F(x);
}

double ThreeBodyIntegrals::ScalarIntegral(double x) {
  return I2(x) + (x - R2 - 2 * sqr(rk)) * I1(x);
}

double ThreeBodyIntegrals::PseudoscalarIntegral(double x) {
  return I2(x) + (x - 1 + sqr(rslj)) * I1(x);
}

double ThreeBodyIntegrals::GaugeIntegral1(double x) {
  return (2 * x - 2 -  sqr(rk)) * I2(x) +
    (2 - (2 + sqr(rk)) * x - 2* sqr(rslj) + 3 * sqr(rk) + sqr(rk)*sqr(rslj)) 
      * I1(x);
}

double ThreeBodyIntegrals::GaugeIntegral2(double x) {
  return sqr(rk) * (I2(x) + (x - 3 - sqr(rslj)) * I1(x));
}

double ThreeBodyIntegrals::FermionFermionCrossIntegral1(double x) {
  return (2 * sqr(rk) * sqr(rm) - x * (1 + sqr(rm)) + R2) * F1(x)
    - (1 - x + sqr(rk)) * F2(x);
}

double ThreeBodyIntegrals::FermionFermionCrossIntegral2(double x) {
  return rk * rm * (1 + sqr(rslj) - sqr(rk) - sqr(rm)) * F1(x);
}

double ThreeBodyIntegrals::FermionFermionCrossIntegral3(double x) {
  return rm * rb * (R2 * F1(x) - F2(x));
}

double ThreeBodyIntegrals::FermionFermionCrossIntegral4(double x) {
  return rk * rb * (F2(x) - 2 * sqr(rm) * F1(x));
}

double ThreeBodyIntegrals::FermionFermionCrossIntegral5(double x) {
  return rm * ra * (x - 2 * sqr(rk)) * F1(x);
}

double ThreeBodyIntegrals::FermionFermionCrossIntegral6(double x) {
  return rk * ra * (R2 - x) * F1(x);
}

double ThreeBodyIntegrals::FermionFermionCrossIntegral7(double x) {
  return ra * rb * ((x - R2) * F1(x) +  F2(x));
}

double ThreeBodyIntegrals::FermionFermionCrossIntegral8(double x) {
  return 2 * rk * rm * ra * rb * F1(x);
}

double ThreeBodyIntegrals::ScalarScalarCrossIntegral(double x) {
  return H2(x) + (x - R2 - 2 * sqr(rk)) * H1(x);
}

double ThreeBodyIntegrals::PseudoscalarPseudoscalarCrossIntegral(double x) {
  return H2(x) + (x - 1 + sqr(rslj)) * H1(x);
}

double ThreeBodyIntegrals::ScalarFermionCrossIntegral1(double x) {
  return rb * (G2(x) + (x - R2 - 2 *  sqr(rk)) * G1(x));
}

double ThreeBodyIntegrals::ScalarFermionCrossIntegral2(double x) {
  return rk * (2 * x - R2 - 2 * sqr(rk)) * G1(x);
}

double ThreeBodyIntegrals::PseudoscalarFermionCrossIntegral1(double x) {
  return rb * (G2(x) + (x - 1 + sqr(rslj)) * G1(x));
}

double ThreeBodyIntegrals::PseudoscalarFermionCrossIntegral2(double x) {
  return rk * (1 - sqr(rslj)) * G1(x);
}

double ThreeBodyIntegrals::ScalarGaugeCrossIntegral(double x) {
  return rk * (H2(x) - x * H1(x));
}

double ThreeBodyIntegrals::A0GaugeCrossIntegral(double x) {
  return rk * (1 - sqr(rslj)) * H1(x);
}

double ThreeBodyIntegrals::G0GaugeCrossIntegral(double x) {
  return rk * (1 - sqr(rslj)) * I1(x);
}

double ThreeBodyIntegrals::GaugeFermionCrossIntegral1(double x) {
  return 2 * (1 - x + sqr(rk)) * G2(x) 
    + (2 * x - 2 * R2 + sqr(rk) * (1 - sqr(rslj))) * G1(x);
}

double ThreeBodyIntegrals::GaugeFermionCrossIntegral2(double x) {
  return rk * rb * (G2(x) - (1 + x - sqr(rslj)) * G1(x));
}

double ThreeBodyIntegrals::GaugeFermionCrossIntegral3(double x) {
  return rk * rb * (G2(x) + (1 - x - sqr(rslj)) * G1(x));
}

double ThreeBodyIntegrals::GaugeFermionCrossIntegral4(double x) {
  return sqr(rk) * (3 - 2 * x + sqr(rslj)) * G1(x);
}

///////////////////////// Simpson's Rule Integrator ///////////////////////////

//Uses Simpsons method to integrate a smooth function from start to end
//(the variables defined in the setMasses).  The method takes as argument
//another method with a single argument of type double.
double ThreeBodyIntegrals::Simpsons(double (*Integral)(double x)) {
  int i;
  int meshmax = 1000;

  //Set the starting and ending location to be slightly within the endpoints
  //This is done to ensure the positivity of the function
  // \lambda (1 - x + rk^2, rm^2, rslj^2), which appears implicitly in every 
  //integral.  The integration region coincides with positive lambda, and 
  //thus at the endpoint lambda vanishes; the endpoints are set to guard 
  //against machine precision driving lambda negative.
  start *= (1. + 1e-12);
  end *= (1. - 1e-12);

  //Set up equally spaced mesh of [meshmax] intervals
  double x[meshmax+1];
  for(i = 0; i < meshmax+1; i++)
    x[i] = start + double(i)*(end-start)/double(meshmax);
  
  double sum = 0.0;

  //Take the weighted sum over the mesh
  for(i = 1; i < meshmax + 1 ; i++) {
    sum += (x[i]-x[i-1])/6. * (Integral(x[i-1]) +
			       4.0*Integral(0.5*(x[i-1]+x[i])) +
			       Integral(x[i]));
  }

  return sum;
}
