#ifndef THREEBODYINTEGRALS_H
#define THREEBODYINTEGRALS_H

#include<softsusy.h>

//Class used to perform the phase-space integration in the slepton three-body
//integrals.
class ThreeBodyIntegrals {
 private:

  //Variable used in the three-body integrals
  static double start, end;
  static double rslj, rk, rm, ra, rb, R2;

  //Helper methods
  static double Lambda(double x, double y, double z);  //Phase space factor
  static double FirstUpperLimit(double x);  //Upper limit of the first integral
  static double FirstLowerLimit(double x);  //Lower limit of the first integral

  //Phase space factors
  //These functions encapsulate the details of the integral over the first
  //fermion energy, including the propagators of the intermediate particles.
  static double K(double x);
  static double L(double x, double z);
  static double F(double x);
  static double F1(double x);
  static double F2(double x);
  static double G1(double x);
  static double G2(double x);
  static double H1(double x);
  static double H2(double x);
  static double I1(double x);
  static double I2(double x);

  //Integrals used for 2 fermion (neutralino or chargino) intermediates
  static double FermionIntegral1(double x);
  static double FermionIntegral2(double x);
  static double FermionIntegral3(double x);
  static double FermionIntegral4(double x);
  static double FermionIntegral5(double x);
  static double FermionIntegral6(double x);

  //Integral used for 2 scalar intermediates of the same type
  static double ScalarIntegral(double x);

  //Integral used for 2 pseudoscalar intermediates of the same type
  static double PseudoscalarIntegral(double x);

  //Integrals used for 2 Z boson intermediates of the same type
  static double GaugeIntegral1(double x);
  static double GaugeIntegral2(double x);

  //Integrals used for 2 neutralino intermediates for the interference term
  //in like-signed lepton decays
  static double FermionFermionCrossIntegral1(double x);
  static double FermionFermionCrossIntegral2(double x);
  static double FermionFermionCrossIntegral3(double x);
  static double FermionFermionCrossIntegral4(double x);
  static double FermionFermionCrossIntegral5(double x);
  static double FermionFermionCrossIntegral6(double x);
  static double FermionFermionCrossIntegral7(double x);
  static double FermionFermionCrossIntegral8(double x);

  //Integral used for 2 scalar intermediates of different types
  static double ScalarScalarCrossIntegral(double x);

  //Integral used for 2 pseudoscalar intermediates of different types
  static double PseudoscalarPseudoscalarCrossIntegral(double x);

  //Integrals used for a neutralino and scalar intermediate 
  static double ScalarFermionCrossIntegral1(double x);
  static double ScalarFermionCrossIntegral2(double x);

  //Integrals used for a neutralino and pseudoscalar intermediates
  static double PseudoscalarFermionCrossIntegral1(double x);
  static double PseudoscalarFermionCrossIntegral2(double x);

  //Integral used for a scalar and a Z boson intermediate
  static double ScalarGaugeCrossIntegral(double x);

  //Integral used for an A0 and a Z boson intermediate
  static double A0GaugeCrossIntegral(double x);

  //Integral used for an G0 and a Z boson intermediate
  static double G0GaugeCrossIntegral(double x);

  //Integrals used for a fermion (neutralino or chargino) and Z intermediate
  static double GaugeFermionCrossIntegral1(double x);
  static double GaugeFermionCrossIntegral2(double x);
  static double GaugeFermionCrossIntegral3(double x);
  static double GaugeFermionCrossIntegral4(double x);

  //  Use Simpsons method to integrate smooth function from start to end
  double Simpsons(double (*Integral)(double x));

 public:

  //Constructor
  ThreeBodyIntegrals();

  //Define the mass parameters used in the integrals
  void setMasses(DoubleVector& masses, double m0);

  //Evaluation methods for the same-type particles
  double evaluateNeutralinoIntegrals(DoubleVector& c);
  double evaluateCharginoIntegral(double c);
  double evaluateScalarIntegral(double c);
  double evaluatePseudoscalarIntegral(double c);
  double evaluateGaugeIntegrals(DoubleVector& c);

  //Evaluation methods for the different-type particles
  double evaluateNeutralinoCrossIntegrals(DoubleVector& c);
  double evaluateScalarScalarCrossIntegral(double c);
  double evaluatePseudoscalarPseudoscalarCrossIntegral(double c);
  double evaluateScalarFermionCrossIntegrals(DoubleVector& c);
  double evaluatePseudoscalarFermionCrossIntegrals(DoubleVector& c);
  double evaluateScalarGaugeCrossIntegral(double c);
  double evaluateA0GaugeCrossIntegral(double c);
  double evaluateG0GaugeCrossIntegral(double c);
  double evaluateGaugeFermionCrossIntegrals(DoubleVector& c);
  double evaluateGaugeCharginoCrossIntegral(double c);
};

#endif
