///////////////////////////////////////////////////////////////////////////////
//                              spice
//                        Main Driver Program
//    This code contains the main routine for running Softsusy.  It is 
//  for both reading input parameters supplied by the user (default input 
//  script file is SPICEinit) and establishing the appropriate flavorful
//  boundary conditions used in renormalization.
//    The program follows these steps.
//       1. User-defined SPICEinit (default script file) is read by
//  readParams.  Information about U(1) charges for sleptons and sneutrinos is
//  used to determine the slepton 6x6 and sneutrino 3x3 mixing matrices.  
//  These mixing matrices are used in flavor-violating boundary conditions 
//  in gravityBcs(...).
//       2. SOFTSUSY uses these boundary conditions and the model parameters 
//  from SPICEinit to derive the low-scale mass hierarchy.  These input
//  parameters can either specify GMSB-, AMSB-, or SUGRA-type models.
//       3. SPICE calculates all relevant decay widths involving 
//  flavor-violating sleptons.  These are sent to FileCombine.x in the file
//  spicydecays.out.  The mass spectrum from SOFTSUSY is sent to SUSYHIT
//  as softsusySLHA.out (which is renamed to slhaspectrum.in per SUSYHIT
//  specifications), and the spectrum is also sent to FileCombine.x as
//  softsusySLHA2.out.
//
///////////////////////////////////////////////////////////////////////////////

#include <fstream>
#include <spice.h>
#include <DecayFormat.h>
#define FLAVORFULL

// start of definition of global variables
// quark mixing (dominant third family approx), and no verboseoutput
int MIXING = 1, PRINTOUT = 0;
// Do we include 2-loop RGEs of *all* scalar masses and A-terms, or only the
// scalar mass Higgs parameters? (Other quantities all 2-loop anyway): the
// default in SOFTSUSY 2.x is to include all 2-loop terms
bool INCLUDE_2_LOOP_SCALAR_CORRECTIONS = true;
// fractional accuracy required
double TOLERANCE = 1.0e-6;
// there are two possible conventions: if QEWSB > MZ, its value is assumed
// in GeV and used as a constant MSUSY. Otherwise, it MULTIPLIES the usual
// MSUSY value, of root(mstop1 mstop2)
double QEWSB = 1.0;
// number of loops used to calculate Higgs mass and tadpoles. They should be
// identical for a consistent calculation
int numRewsbLoops = 2, numHiggsMassLoops = 2;
// decay constant of muon
double GMU = 1.16637e-5;
/// global pole mass of MZ in GeV - MZCENT is defined in def.h
double MZ = MZCENT;
// end of global variable declaration

// Incorrect input: gives advice on how to supply it
void errorCall() {
  ostringstream ii;
  ii << "spice called with incorrect arguments. Need to put:\n";
  ii << "spice.x <n5> <mMess> <lambda> <cgrav> <tanb> <sgnMu> \n";
  throw ii.str();
}

// reads in the parameters of the model from the file and randomly
// sets O(1) parameters. The results are stored in the inParams vector
void readParams(char* fname, DoubleVector &inParams) {
  char inChar[15];
  char dummy[100];

  // common variables
  double tanB;
  int signMu;

  // gmsb variables
  double Lambda, mMess, cgrav;
  int N5;

  // common amsb and sugra variables
  double m0, mGUT;

  // amsb variables
  double m32;

  // sugra variables
  double m12, a0;

  // FV gravity variables
  double x, lam;
  double dRand, coeff;
  int i, j, nCharges, charge, nCount = 1;

  ifstream inFile;
  inFile.open(fname);
  if(!inFile.is_open()) {
    cout << "Error opening file" << endl;
    exit(1);
  }

  try {
    inParams.setEnd(10);
    inFile >> inChar;
    // use strcmp inChar to compare strings
    if(strcmp(inChar,"gmsb") == 0) { // matches gmsb
      inParams.set(1,1); // set first element to be gmsb tag
      inFile >> N5 >> mMess >> Lambda >> cgrav >> tanB >> signMu;
      inParams.set(2,double(N5));
      inParams.set(3,mMess);
      inParams.set(4,Lambda);
      inParams.set(5,cgrav);
      inParams.set(6,tanB);
      inParams.set(7,double(signMu));
    }
    if(strcmp(inChar,"amsb") == 0) { // matches amsb
      inParams.set(1,2); // set first element to be amsb tag
      inFile >> m0 >> m32 >> tanB >> mGUT >> signMu;
      inParams.set(2,m0);
      inParams.set(3,m32);
      inParams.set(4,tanB);
      inParams.set(5,mGUT);
      inParams.set(6,double(signMu));
      inParams.set(7,-9999.0); // placeholder
    }
    if(strcmp(inChar,"sugra") == 0) { // matches sugra
      inParams.set(1,3); // set first element to be sugra tag
      inFile >> m0 >> m12 >> a0 >> tanB >> mGUT >> signMu;
      inParams.set(2,m0);
      inParams.set(3,m12);
      inParams.set(4,a0);
      inParams.set(5,tanB);
      inParams.set(6,mGUT);
      inParams.set(7,double(signMu));
    }
    inFile >> dummy >>  x >> dummy >> lam >> dummy >> nCharges;
    inParams.set(8, x);
    inParams.set(9, lam);
    inParams.set(10, double(nCharges));

// Use a second parameters vector to hold the charges and the
// coefficients

// We take all the matrix elements as inputs even though XL, XR are
// symmetric (just to be more user-friendly)
// Read in one row at a time

// Will have 10 SB params, 6 charge assignments per charge, and 4 matrices of 9
// coefficients each
    inParams.setEnd(10 + nCharges*6 + 4*9);

// Read the charges
    for(i = 0; i < 6; i++) {
      inFile >> dummy; // label of which charged multiplet
      for(j = 0; j < nCharges;j++) {
        inFile >> charge;
        inParams.set(10 + i*nCharges + j + 1, charge);
      }
    }

    inFile >> inChar;
// Generate random numbers for matrices
// First 9 are lepton, then XL, then XR
    if(!strcmp(inChar,"random")) { 
      long rseed;
      double sigma;
      inFile >> rseed >> sigma;
      double* dRand2 = new double [27];
      dRand2 = O1params(rseed, sigma);
      for(i = 0; i < 27; i++)
        inParams.set(10 + nCharges*6 + 1 + i, dRand2[i]);
      delete dRand2;
    }
    else{
// have 9 lepton coefficients, then placeholder, then 9 XL coeff.,
// then placeholder, then 9 XR coeff.
      for(i = 0; i < 27; i++) {
        if(i%9 == 0 && i > 0)
          inFile >> dummy;
        inFile >> coeff;
        inParams.set(10 + nCharges*6 + 1 + i, coeff);
      }
    }
// Make inParams repeat the XL elements to the end of the vector
    for(i = 1; i <= 9; i++)
      inParams.set(10 + nCharges*6 + 3*9 + i, 
                   inParams.display(10 + nCharges*6 + 1*9 + i));
  }
  catch (...) {
    cout << "error" << endl;
    exit(1);
  }
}

// This function generates 21 unique random numbers for use later in
// the XL,XR and leptonic mass matrix in a 27 element array

// First 9 elements are for lepton matrix, next 9 are for XL, last 9
// are for XR
double* O1params(long k, double sigma){
  cout << "Sigma: " << sigma << endl;
  int i;
  double *r = new double [27];
  // Need to use negative k for Softsusy's ran1 function
  if(k > 0) k *= -1;
  if(sigma <= 0.0){
    cout << "Input sigma <= 0 invalid.  Setting sigma = 1.0" << endl;
    sigma = 1.0;
  }
  for(i = 0; i < 27; i++){
    // Softsusy's gasdev fcn returns gaussian distributed random
    // number with mean 0 and variance 1
    r[i] = exp(sigma*gasdev(k));
    if(ran1(k) < 0.5) r[i] *= -1.; // half probability to be neg. coeff.
  }
// Enforce symmetric matrices XL and XR
  r[12] = r[10]; r[15] = r[11]; r[16] = r[14];
  r[21] = r[19]; r[24] = r[20]; r[25] = r[23];

  return r;
}

// Sets the gauge gravity mixed conditions at the high scale. Gives
// the GMSB contributions, then builds the X_L and X_R matrices adding
// O(1) parameters and finally rotates these matrices to the mass
// basis of the charged leptons.
void gaugegravityBcs(MssmSoftsusy &m, const DoubleVector &inParams) {
  DoubleVector vecGmsb(4);
  vecGmsb.set(1,inParams.display(2)); // N5
  vecGmsb.set(2,inParams.display(3)); // mMess
  vecGmsb.set(3,inParams.display(4)); // Lambda
  vecGmsb.set(4,inParams.display(5)); // cgrav
  gmsbBcs(m, vecGmsb);

  gravityBcs(m, inParams);
  return;
}

void anomalygravityBcs(MssmSoftsusy &m, const DoubleVector &inParams) {
  DoubleVector vecAmsb(2);
  vecAmsb.set(1,inParams.display(3)); // m32
  vecAmsb.set(2,inParams.display(2)); // m0

  amsbBcs(m, vecAmsb);

  gravityBcs(m, inParams);
  return;
}

void sugragravityBcs(MssmSoftsusy &m, const DoubleVector &inParams) {
  DoubleVector vecSugra(3);
  vecSugra.set(1,inParams.display(2)); // m0
  vecSugra.set(2,inParams.display(3)); // m12
  vecSugra.set(3,inParams.display(4)); // a0

  sugraBcs(m, vecSugra);

  gravityBcs(m, inParams);
  return;
}

void gravityBcs(MssmSoftsusy &m, const DoubleVector &inParams) {
  double lambda, x, mSquaredGrav;
  DoubleVector W(3);
  DoubleMatrix LepYukawa(3,3), SLepXL(3,3), SLepXR(3,3), mSoft(3,3),
    UL(3,3), UR(3,3);

  x = inParams.display(8);
  BuildLeptonicMatrices(inParams, LepYukawa, SLepXL, SLepXR);

  //Diagonalize the lepton yukawa matrix
  LepYukawa.diagonalise(UL,UR,W);
  UL = UL.transpose();
  UR = UR.transpose();

  // Gravity additions + Rotation to basis in which leptonic Yukawas are
  // diagonal (Superfield Rotation)
  // We do not rotate the trilinear terms because they start at zero at
  // high energy
  mSoft = m.displaySoftMassSquared(mLl);
  mSquaredGrav = x * (mSoft(1,1) + mSoft(2,2) + mSoft(3,3)) / 3;
  mSoft = mSoft + mSquaredGrav * SLepXL;
  mSoft = UL*mSoft*UL.transpose();
  m.setSoftMassMatrix(mLl,mSoft);
  mSoft = m.displaySoftMassSquared(mEr);
  mSoft = mSoft + mSquaredGrav * SLepXR;
  mSoft = UR*mSoft*UR.transpose();
  m.setSoftMassMatrix(mEr,mSoft);

  return;
}

// This function builds the leptonic Yukawa matrix and the sleptonic XL
// and XR matrices using the U(1)charges specified in the
// initialization file.
void BuildLeptonicMatrices(const DoubleVector &inParams, 
                           DoubleMatrix &LepYukawa, DoubleMatrix &SLepXL, 
                           DoubleMatrix &SLepXR){
  int i, j, k, nSum, nSumL, nSumR, nCL, nCR, nCL1, nCL2, nCR1, nCR2, 
      nCharges, nCount;
  bool HolomorphicZero;
  double lambda = inParams.display(9);
  nCharges = (int)inParams.display(10);

  for(i = 0; i < 3; i++)
    for(j = 0; j < 3; j++) {
      nSum = 0;
      nSumL = nSumR = 0;
      HolomorphicZero = false;
      for(k = 1; k <= nCharges; k++) {
        //For the Lepton Yukawa matrix
        nCL = (int)inParams.display(10 + i*nCharges + k);
        nCR = (int)inParams.display(10 + 3*nCharges + j*nCharges + k);
     	if(nCL+nCR < 0)
          HolomorphicZero = true;
        else
          nSum+=nCL+nCR;
        //For The XL, XR matrices
    	nCL1 = (int)inParams.display(10 + i*nCharges + k);
    	nCL2 = (int)inParams.display(10 + j*nCharges + k);
    	nCR1 = (int)inParams.display(10 + 3*nCharges + i*nCharges + k);
    	nCR2 = (int)inParams.display(10 + 3*nCharges + j*nCharges + k);
    	nSumL+=abs(-nCL1+nCL2);
    	nSumR+=abs(-nCR1+nCR2);
      }
      //For the Lepton Yukawa matrix
      if(HolomorphicZero)
        LepYukawa(i+1, j+1) = 0;
      else
        LepYukawa(i+1, j+1) = pow(lambda, nSum)*
            inParams.display(10 + nCharges*6 + i*3 + j + 1);

      //For The XL, XR matrices
      SLepXL(i+1, j+1) = pow(lambda, nSumL)*
          inParams.display(10 + nCharges*6 + 9 + i*3 + j + 1);
      SLepXR(i+1, j+1) = pow(lambda, nSumR)*
          inParams.display(10 + nCharges*6 + 18 + i*3 + j + 1);
    }

  for(i = 1; i <= 3; i++)
    for(j = i+1; j <= 3; j++)
      if(fabs((SLepXL(i,j) - SLepXL(j,i))/SLepXL(i,j)) > 1e-14)
  	cout<<"Warning: Non-diagonal XL flavor-violating matrix"<<endl;

  for(i = 1; i <= 3; i++)
    for(j = i+1; j <= 3; j++)
      if(fabs((SLepXR(i,j) - SLepXR(j,i))/SLepXR(i,j)) > 1e-14)
  	cout<<"Warning: Non-diagonal XR flavor-violating matrix"<<endl;

  return;
}

int main(int argc, char *argv[]) {

  // Sets format of output: 4 decimal places
  outputCharacteristics(6);

  void (*boundaryCondition)(MssmSoftsusy &, const DoubleVector &);

  FlavourMssmSoftsusy m;
  FlavourMssmSoftsusy * r = &m;

  cout << "Beginning SOFTSUSY program" << endl << endl;

  try {
    QedQcd oneset;

    cerr << "SOFTSUSY " << SOFTSUSY_VERSION << endl;
    cerr << "B.C. Allanach, Comput. Phys. Commun. 143 (2002) 305-331,";
    cerr << " hep-ph/0104145\n";
    cerr << "-------------------------------------" << endl;
    cerr << "SPICE" << endl;
    cerr << "Guy Engelhard, Jonathan Feng, Iftah Galon, " 
         << "David Sanford, Felix Yu" << endl;
    cerr << "arXiv:0904.1415v1 [hep-ph]" << endl;

    double mgutGuess = 2.0e16, tanB = 10., mbmb = MBOTTOM, mtau = MTAU;
    int signMu = 1;

    // gmsb variables
    double Lambda=0,cgrav=0,mMess=0;
    // amsb and sugra variables 
    double m0, mGUT, m32, m12, a0;
    bool gaugeUnification = true, ewsbBCscale = false;

    DoubleVector pars(1);
    readParams(argv[1], pars);

    if(pars.display(1) == 1) { // gmsb
      boundaryCondition = &gaugegravityBcs;
      mMess = pars.display(3);
      mgutGuess=mMess;
      Lambda = pars.display(4);
      cgrav = pars.display(5);
      tanB = pars.display(6);
      signMu =  (int)pars.display(7);
      // Error checking on input variables
      if (Lambda > mMess) {
        ostringstream ii;
        ii << "Input Lambda=" << Lambda << " should be greater than mMess="
           << mMess << endl;
        throw ii.str();
      }
      if (cgrav > 1.0) {
        ostringstream ii;
        ii << "Input cgrav=" << cgrav << " a real number bigger than or "
           << " equal to 1 (you can use 1 as a default value).\n";
        throw ii.str();
      }
    }
    if(pars.display(1) == 2) { // amsb
      boundaryCondition = &anomalygravityBcs;
      m0 = pars.display(2);
      m32 = pars.display(3);
      tanB = pars.display(4);
      mGUT = pars.display(5);
      signMu =  (int)pars.display(6);
      mgutGuess = mGUT;
    }
    if(pars.display(1) == 3) { // sugra
      boundaryCondition = &sugragravityBcs;
      m0 = pars.display(2);
      m12 = pars.display(3);
      a0 = pars.display(4);
      tanB = pars.display(5);
      mGUT = pars.display(6);
      signMu =  (int)pars.display(7);
      mgutGuess = mGUT;
    }

    gaugeUnification = false;
    r = &m;
    if (signMu != 1 && signMu != -1 && signMu != 0) {
      ostringstream ii;
      ii << "Incorrect input for sign(mu)=" << signMu <<endl;
      throw ii.str();
    }
    if (tanB < 1.0 || tanB > 1.0e2)  {
      ostringstream ii;
      ii << "Incorrect input for tan beta=" << tanB <<endl;
      throw ii.str();
    }

    oneset.toMz();

    double mgut =  r->lowOrg(boundaryCondition, mgutGuess, pars, signMu,
			     tanB, oneset, gaugeUnification, ewsbBCscale);

    // In order to output the Model's Spectrum at the Higgs
    // Minimization Scale - Msusy which is the scale used as the
    // default option of lesHouchesAccordOutput that is, when
    // specifying qMax=0 and numPoints=1.
    // we run the class to that scale
    mgut = r->runto(r->displayMsusy(), 0.001); 

    // Run to Msusy scale where we'll quote the values for the parameters.

    // Two output files are produces - one in SUSY LesHouches accord format 
    // for use in producing the non-flavor violating decay spectrum by 
    // susyhit, and one in SUSY LesHouches accord 2 format for use in
    // generating the final file.
    ofstream fout, fout2;
    fout.open("softsusySLHA.out");
    fout.setf(ios::scientific, ios::floatfield);
    fout.precision(8);
    fout2.open("softsusySLHA2.out");
    fout2.setf(ios::scientific, ios::floatfield);
    fout2.precision(8);
    if(pars.display(1) == 1){
      DoubleVector vecGmsb(4);
      vecGmsb.set(1, pars.display(2)); // N5
      vecGmsb.set(2, pars.display(3)); // M_mess
      vecGmsb.set(3, pars.display(4)); // Lambda
      vecGmsb.set(4, pars.display(5)); // c_grav
      r->MssmSoftsusy::lesHouchesAccordOutput(fout, "gmsb", vecGmsb, signMu, 
					      tanB, 0,  1, mgutGuess, oneset, 
					      false);
      r->lesHouchesAccordOutput(fout2, "gmsb", vecGmsb, signMu, tanB, 0, 
                                1, mgutGuess, oneset, false);
    }
    if(pars.display(1) == 2){
      DoubleVector vecAmsb(2);
      vecAmsb.set(1, pars.display(3));
      vecAmsb.set(2, pars.display(2));
      r->MssmSoftsusy::lesHouchesAccordOutput(fout, "amsb", vecAmsb, signMu, 
					      tanB, 0, 1, mgutGuess, oneset, 
					      false);
      r->lesHouchesAccordOutput(fout2, "amsb", vecAmsb, signMu, tanB, 0, 
                                1, mgutGuess, oneset, false);
    }
    if(pars.display(1) == 3){
      DoubleVector vecSugra(3);
      vecSugra.set(1, pars.display(2));
      vecSugra.set(2, pars.display(3));
      vecSugra.set(3, pars.display(4));
      r->MssmSoftsusy::lesHouchesAccordOutput(fout, "sugra", vecSugra, signMu, 
					      tanB, 0, 1, mgutGuess, oneset, 
					      false);
      cout<<"Warning: SOFTSUSY Sugra SLHA2 output currently broken."<<endl;
      fout2<<"Warning: SOFTSUSY Sugra SLHA2 output currently broken."<<endl
	   <<"To enable, replace lines 345-347 of flavoursoft.cpp  with"<<endl
	   <<"lines 9217-9219 of softsusy.cpp."<<endl;
      //r->lesHouchesAccordOutput(fout2, "sugra", vecSugra, signMu, tanB, 0, 
      //                          1, mgutGuess, oneset, false);
    }

    cout << "End SOFTSUSY program" << endl;

    // Create a DecayFormat object and use it to generate the 
    // flavor-violating slepton decays
    DecayFormat* DecayOut = new DecayFormat(r);
    DecayOut->OutputIntermediateFile();
  }
  catch(const string &a) {
    cout << a;
  }

  if (r->displayProblem().test()) {
    cout << "# SOFTSUSY problem with point: " << r->displayProblem();
  }

  exit(0);
}
